# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################
import base64
import datetime
import json
import os
from functools import cmp_to_key

from tornado.httputil import url_concat
from tornado.web import authenticated

import sickrage
from sickrage.core.databases.main import MainDB
from sickrage.core.enums import HomeLayout, HistoryLayout, PosterSortBy, PosterSortDirection
from sickrage.core.helpers import remove_article
from sickrage.core.media.util import series_image, SeriesImageType
from sickrage.core.tv.show.coming_episodes import ComingEpisodes, ComingEpsLayout, ComingEpsSortBy
from sickrage.core.tv.show.helpers import get_show_list, find_show
from sickrage.core.webserver.handlers.api.v1 import ApiV1Handler
from sickrage.core.webserver.handlers.base import BaseHandler


class RobotsDotTxtHandler(BaseHandler):
    def initialize(self):
        self.set_header('Content-Type', 'text/plain')

    def get(self, *args, **kwargs):
        """ Keep web crawlers out """
        return "User-agent: *\nDisallow: /"


class MessagesDotPoHandler(BaseHandler):
    def initialize(self):
        self.set_header('Content-Type', 'text/plain')

    @authenticated
    def get(self, *args, **kwargs):
        """ Get /sickrage/locale/{lang_code}/LC_MESSAGES/messages.po """
        if sickrage.app.config.gui.gui_lang:
            locale_file = os.path.join(sickrage.LOCALE_DIR, sickrage.app.config.gui.gui_lang, 'LC_MESSAGES/messages.po')
            if os.path.isfile(locale_file):
                with open(locale_file, 'r', encoding='utf8') as f:
                    return f.read()

class APIBulderHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        def titler(x):
            return (remove_article(x), x)[not x or sickrage.app.config.general.sort_article]

        episodes = {}

        for show_object in get_show_list():
            if show_object.series_id not in episodes:
                episodes[show_object.series_id] = {}

            for episode_object in show_object.episodes:
                if episode_object.season not in episodes[show_object.series_id]:
                    episodes[show_object.series_id][episode_object.season] = []

                episodes[show_object.series_id][episode_object.season].append(episode_object.episode)

        if len(sickrage.app.config.general.api_v1_key) == 32:
            apikey = sickrage.app.config.general.api_v1_key
        else:
            apikey = _('API Key not generated')

        api_commands = {}
        for command, api_call in ApiV1Handler(self.application, self.request).api_calls.items():
            api_commands[command] = api_call(self.application, self.request, **{'help': 1}).run()

        return self.render('api_builder.mako',
                           title=_('API Builder'),
                           header=_('API Builder'),
                           shows=sorted(get_show_list(), key=cmp_to_key(lambda x, y: titler(x.name) < titler(y.name))),
                           episodes=base64.b64encode(json.dumps(episodes).encode()).decode(),
                           apikey=apikey,
                           api_commands=api_commands,
                           controller='root',
                           action='api_builder')


class SetHomeLayoutHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        layout = self.get_argument('layout', 'POSTER')

        if layout not in ('POSTER', 'SMALL', 'BANNER', 'SIMPLE', 'DETAILED'):
            layout = 'POSTER'

        sickrage.app.config.gui.home_layout = HomeLayout[layout]
        sickrage.app.config.save()

        # Don't redirect to default page so user can see new layout
        return self.redirect("/home/")


class SetPosterSortByHandler(BaseHandler):
    @authenticated
    def post(self, *args, **kwargs):
        sort = self.get_argument('sort', 'NAME')

        if sort not in ('NAME', 'DATE', 'NETWORK', 'PROGRESS'):
            sort = 'NAME'

        sickrage.app.config.gui.poster_sort_by = PosterSortBy[sort]
        sickrage.app.config.save()


class SetPosterSortDirHandler(BaseHandler):
    @authenticated
    def post(self, *args, **kwargs):
        direction = self.get_argument('direction', 'ASCENDING')

        sickrage.app.config.gui.poster_sort_dir = PosterSortDirection[direction]
        sickrage.app.config.save()


class SetHistoryLayoutHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        layout = self.get_argument('layout', 'DETAILED')

        if layout not in ('COMPACT', 'DETAILED'):
            layout = 'DETAILED'

        sickrage.app.config.gui.history_layout = HistoryLayout[layout]

        return self.redirect("/history/")


class ToggleDisplayShowSpecialsHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        show = self.get_argument('show')

        sickrage.app.config.gui.display_show_specials = not sickrage.app.config.gui.display_show_specials
        return self.redirect(url_concat("/home/displayShow", {'show': show}))


class SetScheduleLayoutHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        layout = self.get_argument('layout', 'BANNER')

        if layout not in ('POSTER', 'BANNER', 'LIST', 'CALENDAR'):
            layout = 'BANNER'

        if layout == 'CALENDAR':
            sickrage.app.config.gui.coming_eps_sort = ComingEpsSortBy.DATE

        sickrage.app.config.gui.coming_eps_layout = ComingEpsLayout[layout]

        return self.redirect("/schedule/")


class ToggleScheduleDisplayPausedHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        sickrage.app.config.gui.coming_eps_display_paused = not sickrage.app.config.gui.coming_eps_display_paused
        self.redirect("/schedule/")


class SetScheduleSortHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        sort = self.get_argument('sort', 'DATE')

        if sort not in ('DATE', 'NETWORK', 'SHOW'):
            sort = 'DATE'

        if sickrage.app.config.gui.coming_eps_layout == ComingEpsLayout.CALENDAR:
            sort = 'DATE'

        sickrage.app.config.gui.coming_eps_sort = ComingEpsSortBy[sort]

        return self.redirect("/schedule/")


class ScheduleHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        layout = self.get_argument('layout', None)

        next_week = datetime.datetime.combine(datetime.date.today() + datetime.timedelta(days=7),
                                              datetime.datetime.now().time().replace(tzinfo=sickrage.app.tz))

        today = datetime.datetime.now().replace(tzinfo=sickrage.app.tz)

        results = ComingEpisodes.get_coming_episodes(ComingEpisodes.categories, sickrage.app.config.gui.coming_eps_sort, False)

        return self.render('schedule.mako',
                           next_week=next_week,
                           today=today,
                           results=results,
                           layout=ComingEpsLayout[layout] if layout else sickrage.app.config.gui.coming_eps_layout,
                           title=_('Schedule'),
                           header=_('Schedule'),
                           topmenu='schedule',
                           controller='root',
                           action='schedule')


class QuicksearchDotJsonHandler(BaseHandler):
    @authenticated
    def post(self, *args, **kwargs):
        term = self.get_argument('term')

        shows = []
        episodes = []

        session = sickrage.app.main_db.session()

        for result in session.query(MainDB.TVShow).filter(MainDB.TVShow.name.like('%{}%'.format(term))).all():
            shows.append({
                'category': 'shows',
                'series_id': result.series_id,
                'series_provider_id': result.series_provider_id.name,
                'seasons': len(set([s.season for s in result.episodes])),
                'name': result.name,
                'img': sickrage.app.config.general.web_root + series_image(result.series_id, result.series_provider_id, SeriesImageType.POSTER_THUMB).url
            })

        for result in session.query(MainDB.TVEpisode).filter(MainDB.TVEpisode.name.like('%{}%'.format(term))).all():
            show_object = find_show(result.series_id, result.series_provider_id)
            if not show_object:
                continue

            episodes.append({
                'category': 'episodes',
                'series_id': result.series_id,
                'series_provider_id': result.series_provider_id.name,
                'episode_id': result.episode_id,
                'season': result.season,
                'episode': result.episode,
                'name': result.name,
                'show_name': show_object.name,
                'img': sickrage.app.config.general.web_root + series_image(result.series_id, result.series_provider_id, SeriesImageType.POSTER_THUMB).url
            })

        if not len(shows):
            shows = [{
                'category': 'shows',
                'series_id': '',
                'series_provider_id': '',
                'name': term,
                'img': '/images/poster-thumb.png',
                'seasons': 0,
            }]

        return json.dumps(shows + episodes)


class ForceSchedulerJobHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        name = self.get_argument('name')

        service = getattr(sickrage.app, name, None)
        if service:
            job = sickrage.app.scheduler.get_job(service.name)
            if job:
                job.modify(next_run_time=datetime.datetime.utcnow(), kwargs={'force': True})



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################

import sentry_sdk
from jose import ExpiredSignatureError, JWTError

import sickrage
from sickrage.core.enums import UserPermission
from sickrage.core.helpers import is_ip_whitelisted, get_internal_ip
from sickrage.core.webserver.handlers.base import BaseHandler


class LoginHandler(BaseHandler):
    def get(self, *args, **kwargs):
        if is_ip_whitelisted(self.request.remote_ip):
            return self.redirect("{}".format(self.get_argument('next', "/{}/".format(sickrage.app.config.general.default_page.value))))
        elif 'Authorization' in self.request.headers:
            return self.handle_jwt_auth_get()
        elif sickrage.app.config.general.sso_auth_enabled and sickrage.app.auth_server.health:
            return self.handle_sso_auth_get()
        elif sickrage.app.config.general.local_auth_enabled:
            return self.handle_local_auth_get()
        else:
            return self.render('login_failed.mako',
                               topmenu="system",
                               header="SiCKRAGE Login Failed",
                               title="SiCKRAGE Login Failed",
                               controller='root',
                               action='login')

    def post(self, *args, **kwargs):
        if sickrage.app.config.general.local_auth_enabled:
            return self.handle_local_auth_post()

    def handle_jwt_auth_get(self):
        certs = sickrage.app.auth_server.certs()
        auth_token = self.request.headers['Authorization'].strip('Bearer').strip()

        try:
            decoded_token = sickrage.app.auth_server.decode_token(auth_token, certs)
        except ExpiredSignatureError:
            self.set_status(401)
            return {'error': 'Token expired'}
        except JWTError as e:
            self.set_status(401)
            return {'error': f'Improper JWT token supplied, {e!r}'}

        if not sickrage.app.config.user.sub_id:
            sickrage.app.config.user.sub_id = decoded_token.get('sub')
            sickrage.app.config.save(mark_dirty=True)

        if sickrage.app.config.user.sub_id == decoded_token.get('sub'):
            save_config = False
            if not sickrage.app.config.user.username:
                sickrage.app.config.user.username = decoded_token.get('preferred_username')
                save_config = True

            if not sickrage.app.config.user.email:
                sickrage.app.config.user.email = decoded_token.get('email')
                save_config = True

            if not sickrage.app.config.user.permissions == UserPermission.SUPERUSER:
                sickrage.app.config.user.permissions = UserPermission.SUPERUSER
                save_config = True

            if save_config:
                sickrage.app.config.save()

        if sickrage.app.config.user.sub_id == decoded_token.get('sub'):
            sentry_sdk.set_user({
                'id': sickrage.app.config.user.sub_id,
                'username': sickrage.app.config.user.username,
                'email': sickrage.app.config.user.email
            })

        if sickrage.app.config.user.sub_id != decoded_token.get('sub'):
            return

        if not sickrage.app.api.token:
            exchanged_token = sickrage.app.auth_server.token_exchange(auth_token)
            if exchanged_token:
                sickrage.app.api.token = exchanged_token

        if not sickrage.app.config.general.server_id:
            server_id = sickrage.app.api.server.register_server(
                ip_addresses=','.join([get_internal_ip()]),
                web_protocol=self.request.protocol,
                web_port=sickrage.app.config.general.web_port,
                web_root=sickrage.app.config.general.web_root,
                server_version=sickrage.version()
            )

            if server_id:
                sickrage.app.config.general.server_id = server_id
                sickrage.app.config.save()

        if sickrage.app.config.general.server_id:
            sentry_sdk.set_tag('server_id', sickrage.app.config.general.server_id)

    def handle_sso_auth_get(self):
        code = self.get_argument('code', None)

        redirect_uri = f"{self.request.protocol}://{self.request.host}{sickrage.app.config.general.web_root}/login"

        if code:
            try:
                token = sickrage.app.auth_server.authorization_code(code, redirect_uri)
                if not token:
                    return self.redirect('/logout')

                certs = sickrage.app.auth_server.certs()
                if not certs:
                    return self.redirect('/logout')

                decoded_token = sickrage.app.auth_server.decode_token(token['access_token'], certs)
                if not decoded_token:
                    return self.redirect('/logout')

                if not decoded_token.get('sub'):
                    return self.redirect('/logout')

                self.set_secure_cookie('_sr_access_token', token['access_token'])
                self.set_secure_cookie('_sr_refresh_token', token['refresh_token'])

                if not sickrage.app.config.user.sub_id:
                    sickrage.app.config.user.sub_id = decoded_token.get('sub')
                    sickrage.app.config.save(mark_dirty=True)

                if sickrage.app.config.user.sub_id == decoded_token.get('sub'):
                    save_config = False
                    if not sickrage.app.config.user.username:
                        sickrage.app.config.user.username = decoded_token.get('preferred_username')
                        save_config = True

                    if not sickrage.app.config.user.email:
                        sickrage.app.config.user.email = decoded_token.get('email')
                        save_config = True

                    if not sickrage.app.config.user.permissions == UserPermission.SUPERUSER:
                        sickrage.app.config.user.permissions = UserPermission.SUPERUSER
                        save_config = True

                    if save_config:
                        sickrage.app.config.save()

                if sickrage.app.config.user.sub_id == decoded_token.get('sub'):
                    sentry_sdk.set_user({
                        'id': sickrage.app.config.user.sub_id,
                        'username': sickrage.app.config.user.username,
                        'email': sickrage.app.config.user.email
                    })

                if sickrage.app.config.user.sub_id != decoded_token.get('sub'):
                    if sickrage.app.api.token:
                        allowed_usernames = sickrage.app.api.allowed_usernames()['data']
                        if not decoded_token.get('preferred_username') in allowed_usernames:
                            sickrage.app.log.debug(
                                "USERNAME:{} IP:{} - WEB-UI ACCESS DENIED".format(decoded_token.get('preferred_username'), self.request.remote_ip))
                            return self.redirect('/logout')
                    else:
                        return self.redirect('/logout')
                elif not sickrage.app.api.token:
                    exchanged_token = sickrage.app.auth_server.token_exchange(token['access_token'])
                    if exchanged_token:
                        sickrage.app.api.token = exchanged_token
            except Exception as e:
                sickrage.app.log.debug('{!r}'.format(e))
                return self.redirect('/logout')

            if not sickrage.app.config.general.server_id:
                server_id = sickrage.app.api.server.register_server(
                    ip_addresses=','.join([get_internal_ip()]),
                    web_protocol=self.request.protocol,
                    web_port=sickrage.app.config.general.web_port,
                    web_root=sickrage.app.config.general.web_root,
                    server_version=sickrage.version()
                )

                if server_id:
                    sickrage.app.config.general.server_id = server_id
                    sickrage.app.config.save()

            if sickrage.app.config.general.server_id:
                sentry_sdk.set_tag('server_id', sickrage.app.config.general.server_id)

            redirect_uri = self.get_argument('next', "/{}/".format(sickrage.app.config.general.default_page.value))
            return self.redirect("{}".format(redirect_uri))
        else:
            authorization_url = sickrage.app.auth_server.authorization_url(redirect_uri=redirect_uri, scope="profile email")
            if authorization_url:
                return self.redirect(authorization_url, add_web_root=False)

        return self.redirect('/logout')

    def handle_local_auth_get(self):
        return self.render('login.mako',
                           topmenu="system",
                           header="SiCKRAGE Login",
                           title="SiCKRAGE Login",
                           controller='root',
                           action='login')

    def handle_local_auth_post(self):
        username = self.get_argument('username', '')
        password = self.get_argument('password', '')
        remember_me = self.get_argument('remember_me', None)

        if username == sickrage.app.config.user.username and password == sickrage.app.config.user.password:
            self.set_secure_cookie('_sr', sickrage.app.config.general.api_v1_key, expires_days=30 if remember_me else 1)
            return self.redirect("{}".format(self.get_argument('next', "/{}/".format(sickrage.app.config.general.default_page.value))))

        return self.redirect("/login")



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
# Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.


import sickrage
from trakt import Trakt


class TraktAPI(object):
    def __init__(self):
        # Set trakt app id
        Trakt.configuration.defaults.app(
            id=sickrage.app.trakt_app_id
        )

        # Set trakt client id/secret
        Trakt.configuration.defaults.client(
            id=sickrage.app.trakt_api_key,
            secret=sickrage.app.trakt_api_secret
        )

        # Bind trakt events
        Trakt.on('oauth.token_refreshed', self.on_token_refreshed)

        Trakt.configuration.defaults.oauth(
            refresh=True
        )

        if sickrage.app.config.trakt.oauth_token:
            Trakt.configuration.defaults.oauth.from_response(
                sickrage.app.config.trakt.oauth_token
            )

    @staticmethod
    def authenticate(pin):
        # Exchange `code` for `access_token`
        sickrage.app.config.trakt.oauth_token = Trakt['oauth'].token_exchange(pin, 'urn:ietf:wg:oauth:2.0:oob')
        if not sickrage.app.config.trakt.oauth_token:
            return False

        sickrage.app.log.debug('Token exchanged - auth: %r' % sickrage.app.config.trakt.oauth_token)
        sickrage.app.config.save()

        return True

    @staticmethod
    def on_token_refreshed(response):
        # OAuth token refreshed, save token for future calls
        sickrage.app.config.trakt.oauth_token = response

        sickrage.app.log.debug('Token refreshed - auth: %r' % sickrage.app.config.trakt.oauth_token)
        sickrage.app.config.save()

    def __getattr__(self, name):
        if hasattr(self, name):
            return super(TraktAPI, self).__getattribute__(name)

        return getattr(Trakt, name)

    def __setattr__(self, name, value):
        if hasattr(self, name):
            return super(TraktAPI, self).__setattr__(name, value)

        setattr(Trakt, name, value)

    def __getitem__(self, key):
        return Trakt[key]


class TraktException(Exception):
    pass


class TraktAuthException(TraktException):
    pass


class TraktServerBusy(TraktException):
    pass



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################


import sickrage
from sickrage.core.webserver.handlers.base import BaseHandler


class LogoutHandler(BaseHandler):
    def get(self, *args, **kwargs):
        logout_uri = sickrage.app.auth_server.get_url('end_session_endpoint') if sickrage.app.config.general.sso_auth_enabled else ""
        redirect_uri = f"{self.request.protocol}://{self.request.host}{sickrage.app.config.general.web_root}/login"

        self.clear_cookie('_sr')
        self.clear_cookie('_sr_access_token')
        self.clear_cookie('_sr_refresh_token')

        if logout_uri:
            return self.redirect(f'{logout_uri}?redirect_uri={redirect_uri}', add_web_root=False)
        else:
            return self.redirect(f'{redirect_uri}', add_web_root=False)



"""Initial migration

Revision ID: 4
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""
import json
import os
from json import JSONDecodeError

import sqlalchemy as sa
from alembic import op

import sickrage

# revision identifiers, used by Alembic.
revision = '4'
down_revision = '3'


def upgrade():
    conn = op.get_bind()
    meta = sa.MetaData(bind=conn)
    oauth2_token = sa.Table('oauth2_token', meta, autoload=True)

    token_file = os.path.abspath(os.path.join(sickrage.app.data_dir, 'token.json'))

    if os.path.exists(token_file):
        with open(token_file, 'r') as fd:
            try:
                token = json.load(fd)
                conn.execute(oauth2_token.insert().values(
                    access_token=token['access_token'],
                    refresh_token=token['refresh_token'],
                    expires_in=token['expires_in'],
                    expires_at=token['expires_at'],
                    scope=' '.join(token['scope']) if isinstance(token['scope'], list) else token['scope']
                ))
            except JSONDecodeError:
                pass

        os.remove(token_file)


def downgrade():
    # Operations to reverse the above upgrade go here.
    pass



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################

import os
import re
from urllib.parse import unquote_plus, urlencode

from tornado.escape import json_encode
from tornado.httputil import url_concat
from tornado.web import authenticated

import sickrage
from sickrage.core.common import Quality, Qualities, EpisodeStatus
from sickrage.core.enums import SearchFormat, SeriesProviderID
from sickrage.core.helpers import sanitize_file_name, make_dir, chmod_as_parent, checkbox_to_value
from sickrage.core.helpers.anidb import short_group_names
from sickrage.core.imdb_popular import imdbPopular
from sickrage.core.traktapi import TraktAPI
from sickrage.core.tv.show import TVShow
from sickrage.core.tv.show.helpers import find_show, get_show_list, find_show_by_location
from sickrage.core.webserver.handlers.base import BaseHandler
from sickrage.series_providers.helpers import search_series_provider_for_series_id


def split_extra_show(extra_show):
    if not extra_show:
        return None, None, None, None

    split_vals = extra_show.split('|')

    if len(split_vals) < 4:
        series_provider_id = split_vals[0]
        show_dir = split_vals[1]
        return series_provider_id, show_dir, None, None

    series_provider_id = split_vals[0]
    show_dir = split_vals[1]
    series_id = split_vals[2]
    show_name = '|'.join(split_vals[3:])

    return series_provider_id, show_dir, series_id, show_name


class HomeAddShowsHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        return self.render('home/add_shows.mako',
                           title=_('Add Shows'),
                           header=_('Add Shows'),
                           topmenu='home',
                           controller='home',
                           action='add_shows')


class SearchSeriesProviderForShowNameHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        search_term = self.get_argument('search_term')
        series_provider_id = self.get_argument('series_provider_id', None)
        lang = self.get_argument('lang', None)

        results = []

        series_provider = sickrage.app.series_providers[SeriesProviderID[series_provider_id]]
        series_provider_language = lang if not lang or lang == 'null' else sickrage.app.config.general.series_provider_default_language

        sickrage.app.log.debug(f"Searching for Show with term: {search_term} on series provider: {series_provider.name}")

        # search via series name
        series_results = series_provider.search(search_term, language=series_provider_language)
        if series_results:
            for series in series_results:
                if not series.get('seriesname', None):
                    continue

                if not series.get('firstaired', None):
                    continue

                results.append((
                    series_provider.name,
                    series_provider_id,
                    series_provider.show_url,
                    int(series['id']),
                    series['seriesname'],
                    series['firstaired'],
                    ('', 'disabled')[isinstance(find_show(int(series['id']), SeriesProviderID[series_provider_id]), TVShow)]
                ))

        return json_encode({'results': results, 'langid': lang})


class MassAddTableHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        root_dir = self.get_arguments('rootDir')

        root_dirs = [unquote_plus(x) for x in root_dir]

        if sickrage.app.config.general.root_dirs:
            default_index = int(sickrage.app.config.general.root_dirs.split('|')[0])
        else:
            default_index = 0

        if len(root_dirs) > default_index:
            tmp = root_dirs[default_index]
            if tmp in root_dirs:
                root_dirs.remove(tmp)
                root_dirs = [tmp] + root_dirs

        dir_list = []

        for root_dir in root_dirs:
            try:
                file_list = os.listdir(root_dir)
            except Exception:
                continue

            for cur_file in file_list:
                try:
                    cur_path = os.path.normpath(os.path.join(root_dir, cur_file))
                    if not os.path.isdir(cur_path):
                        continue

                    # ignore Synology folders
                    if cur_file.lower() in ['#recycle', '@eadir']:
                        continue

                    cur_dir = {'dir': cur_path, 'display_dir': '<b>{}{}</b>{}'.format(os.path.dirname(cur_path), os.sep, os.path.basename(cur_path))}

                    # see if the folder is in database already
                    cur_dir['added_already'] = False
                    if find_show_by_location(cur_path):
                        cur_dir['added_already'] = True

                    dir_list.append(cur_dir)

                    series_id = show_name = series_provider_id = None
                    for cur_provider in sickrage.app.metadata_providers.values():
                        if all([series_id, show_name, series_provider_id]):
                            continue

                        (series_id, show_name, series_provider_id) = cur_provider.retrieve_show_metadata(cur_path)
                        if show_name:
                            if not series_provider_id and series_id:
                                for series_provider_id in SeriesProviderID:
                                    result = search_series_provider_for_series_id(series_provider_id, show_name)
                                    if result == series_id:
                                        break
                            elif not series_id and series_provider_id:
                                series_id = search_series_provider_for_series_id(series_provider_id, show_name)

                    cur_dir['existing_info'] = (series_id, show_name, series_provider_id)
                    if series_id and find_show(series_id, series_provider_id):
                        cur_dir['added_already'] = True
                except Exception:
                    pass

        return self.render('home/mass_add_table.mako',
                           dirList=dir_list,
                           controller='home',
                           action="mass_add_table")


class NewShowHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        """
        Display the new show page which collects a tvdb id, folder, and extra options and
        posts them to addNewShow
        """

        show_to_add = self.get_argument('show_to_add', None)
        other_shows = self.get_arguments('other_shows')
        search_string = self.get_argument('search_string', None)

        series_provider_id, show_dir, series_id, show_name = split_extra_show(show_to_add)

        use_provided_info = False
        if series_id and series_provider_id and show_name:
            use_provided_info = True

        # use the given show_dir for the series provider search if available
        default_show_name = show_name or ''
        if not show_dir and search_string:
            default_show_name = search_string
        elif not show_name and show_dir:
            default_show_name = re.sub(r' \(\d{4}\)', '', os.path.basename(os.path.normpath(show_dir)).replace('.', ' '))

        provided_series_id = int(series_id or 0)
        provided_series_name = show_name or ''
        provided_series_provider_id = SeriesProviderID[series_provider_id] if series_provider_id else sickrage.app.config.general.series_provider_default

        return self.render('home/new_show.mako',
                           enable_anime_options=True,
                           use_provided_info=use_provided_info,
                           default_show_name=default_show_name,
                           other_shows=other_shows,
                           provided_show_dir=show_dir,
                           provided_series_id=provided_series_id,
                           provided_series_name=provided_series_name,
                           provided_series_provider_id=provided_series_provider_id,
                           series_providers=SeriesProviderID,
                           quality=sickrage.app.config.general.quality_default,
                           whitelist=[],
                           blacklist=[],
                           groups=[],
                           title=_('New Show'),
                           header=_('New Show'),
                           topmenu='home',
                           controller='home',
                           action="new_show")


class TraktShowsHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        """
        Display the new show page which collects a tvdb id, folder, and extra options and
        posts them to addNewShow
        """

        show_list = self.get_argument('list', 'trending')
        limit = self.get_argument('limit', None) or 10

        trakt_shows = []

        shows, black_list = getattr(TraktAPI()['shows'], show_list)(extended="full", limit=int(limit) + len(get_show_list())), False

        while len(trakt_shows) < int(limit):
            trakt_shows += [x for x in shows if 'tvdb' in x.ids and not find_show(int(x.ids['tvdb']))]

        return self.render('home/trakt_shows.mako',
                           title="Trakt {} Shows".format(show_list.capitalize()),
                           header="Trakt {} Shows".format(show_list.capitalize()),
                           enable_anime_options=False,
                           black_list=black_list,
                           trakt_shows=trakt_shows[:int(limit)],
                           trakt_list=show_list,
                           limit=limit,
                           controller='home',
                           action="trakt_shows")


class PopularShowsHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        """
        Fetches data from IMDB to show a list of popular shows.
        """
        imdb_exception = None

        try:
            popular_shows = imdbPopular().fetch_popular_shows()
        except Exception as e:
            popular_shows = None
            imdb_exception = e

        return self.render('home/imdb_shows.mako',
                           title="IMDB Popular Shows",
                           header="IMDB Popular Shows",
                           popular_shows=popular_shows,
                           imdb_exception=imdb_exception,
                           topmenu="home",
                           controller='home',
                           action="popular_shows")


class AddShowToBlacklistHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        series_id = self.get_argument('series_id')

        data = {'shows': [{'ids': {'tvdb': series_id}}]}
        TraktAPI()["users/me/lists/{list}".format(list=sickrage.app.config.trakt.blacklist_name)].add(data)
        return self.redirect('/home/addShows/trendingShows/')


class ExistingShowsHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        """
        Prints out the page to add existing shows from a root dir
        """
        return self.render('home/add_existing_shows.mako',
                           enable_anime_options=False,
                           quality=sickrage.app.config.general.quality_default,
                           title=_('Existing Show'),
                           header=_('Existing Show'),
                           topmenu="home",
                           controller='home',
                           action="add_existing_shows")


class AddShowByIDHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        series_id = self.get_argument('series_id')
        show_name = self.get_argument('showName')

        if re.search(r'tt\d+', series_id):
            result = sickrage.app.series_providers[SeriesProviderID.THETVDB].search(series_id)
            if result and 'id' in result:
                series_id = int(result['id'])

        if find_show(int(series_id), SeriesProviderID.THETVDB):
            sickrage.app.log.debug(f"{series_id} already exists in your show library, skipping!")
            return

        location = None
        if sickrage.app.config.general.root_dirs:
            root_dirs = sickrage.app.config.general.root_dirs.split('|')
            location = root_dirs[int(root_dirs[0]) + 1]

        if not location:
            sickrage.app.log.warning("There was an error creating the show, no root directory setting found")
            return _('No root directories setup, please go back and add one.')

        show_dir = os.path.join(location, sanitize_file_name(show_name))

        return self.redirect(url_concat("/home/addShows/newShow",
                                        {'show_to_add': '{series_provider_id}|{show_dir}|{series_id}|{show_name}'.format(
                                            **{'series_provider_id': SeriesProviderID.THETVDB.name,
                                               'show_dir': show_dir,
                                               'series_id': series_id,
                                               'show_name': show_name})}))


class AddNewShowHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        return self.redirect("/home/")

    @authenticated
    def post(self, *args, **kwargs):
        """
        Receive tvdb id, dir, and other options and create a show from them. If extra show dirs are
        provided then it forwards back to newShow, if not it goes to /home.
        """

        whichSeries = self.get_argument('whichSeries', None)
        rootDir = self.get_argument('rootDir', None)
        fullShowPath = self.get_argument('fullShowPath', None)
        provided_series_name = self.get_argument('providedSeriesName', None)
        series_provider_language = self.get_argument('seriesProviderLanguage', None)
        defaultStatus = self.get_argument('defaultStatus', None)
        quality_preset = self.get_argument('quality_preset', None)
        anyQualities = self.get_arguments('anyQualities')
        bestQualities = self.get_arguments('bestQualities')
        flatten_folders = self.get_argument('flatten_folders', None)
        subtitles = self.get_argument('subtitles', None)
        sub_use_sr_metadata = self.get_argument('sub_use_sr_metadata', None)
        other_shows = self.get_arguments('other_shows')
        skipShow = self.get_argument('skipShow', None)
        provided_series_provider_id = self.get_argument('providedSeriesProviderID', None)
        anime = self.get_argument('anime', None)
        search_format = self.get_argument('search_format', None)
        dvd_order = self.get_argument('dvd_order', None)
        blacklist = self.get_argument('blacklist', None)
        whitelist = self.get_argument('whitelist', None)
        defaultStatusAfter = self.get_argument('defaultStatusAfter', None)
        scene = self.get_argument('scene', None)
        skip_downloaded = self.get_argument('skip_downloaded', None)
        add_show_year = self.get_argument('add_show_year', None)

        # if we're skipping then behave accordingly
        if skipShow:
            return self.finish_add_show(other_shows)

        if not whichSeries:
            return self.redirect("/home/")

        # figure out what show we're adding and where
        series_pieces = whichSeries.split('|')
        if (whichSeries and rootDir or whichSeries and fullShowPath) and len(series_pieces) > 1:
            if len(series_pieces) < 6:
                sickrage.app.log.error('Unable to add show due to show selection. Not anough arguments: %s' % (repr(series_pieces)))
                sickrage.app.alerts.error(_('Unknown error. Unable to add show due to problem with show selection.'))
                return self.redirect('/home/addShows/existingShows/')

            series_provider_id = series_pieces[1]
            series_id = int(series_pieces[3])
            show_name = series_pieces[4]
        else:
            series_provider_id = provided_series_provider_id or sickrage.app.config.general.series_provider_default
            series_id = int(whichSeries)
            if fullShowPath:
                show_name = os.path.basename(os.path.normpath(fullShowPath))
            else:
                show_name = provided_series_name

        # use the whole path if it's given, or else append the show name to the root dir to get the full show path
        if fullShowPath:
            show_dir = os.path.normpath(fullShowPath)
        else:
            show_dir = os.path.join(rootDir, sanitize_file_name(show_name))
            if add_show_year and not re.match(r'.*\(\d+\)$', show_dir) and re.search(r'\d{4}', series_pieces[5]):
                show_dir = "{} ({})".format(show_dir, re.search(r'\d{4}', series_pieces[5]).group(0))

        # blanket policy - if the dir exists you should have used "add existing show" numbnuts
        if os.path.isdir(show_dir) and not fullShowPath:
            sickrage.app.alerts.error(_("Unable to add show"),
                                      _("Folder ") + show_dir + _(" exists already"))
            return self.redirect('/home/addShows/existingShows/')

        # don't create show dir if config says not to
        if sickrage.app.config.general.add_shows_wo_dir:
            sickrage.app.log.info("Skipping initial creation of " + show_dir + " due to SiCKRAGE configuation setting")
        else:
            dir_exists = make_dir(show_dir)
            if not dir_exists:
                sickrage.app.log.warning("Unable to create the folder " + show_dir + ", can't add the show")
                sickrage.app.alerts.error(_("Unable to add show"),
                                          _("Unable to create the folder " +
                                            show_dir + ", can't add the show"))

                # Don't redirect to default page because user wants to see the new show
                return self.redirect("/home/")
            else:
                chmod_as_parent(show_dir)

        if whitelist:
            whitelist = short_group_names(whitelist)
        if blacklist:
            blacklist = short_group_names(blacklist)

        try:
            new_quality = Qualities[quality_preset.upper()]
        except (AttributeError, KeyError):
            new_quality = Quality.combine_qualities([Qualities[x.upper()] for x in anyQualities], [Qualities[x.upper()] for x in bestQualities])

        # add the show
        sickrage.app.show_queue.add_show(series_provider_id=SeriesProviderID[series_provider_id],
                                         series_id=series_id,
                                         showDir=show_dir,
                                         default_status=EpisodeStatus[defaultStatus],
                                         default_status_after=EpisodeStatus[defaultStatusAfter],
                                         quality=new_quality,
                                         flatten_folders=checkbox_to_value(flatten_folders),
                                         lang=series_provider_language or sickrage.app.config.general.series_provider_default_language,
                                         subtitles=checkbox_to_value(subtitles),
                                         sub_use_sr_metadata=checkbox_to_value(sub_use_sr_metadata),
                                         anime=checkbox_to_value(anime),
                                         dvd_order=checkbox_to_value(dvd_order),
                                         search_format=SearchFormat[search_format],
                                         paused=False,
                                         blacklist=blacklist,
                                         whitelist=whitelist,
                                         scene=checkbox_to_value(scene),
                                         skip_downloaded=checkbox_to_value(skip_downloaded))

        sickrage.app.alerts.message(_('Adding Show'), _('Adding the specified show into ') + show_dir)

        return self.finish_add_show(other_shows)

    def finish_add_show(self, other_shows):
        # if there are no extra shows then go home
        if not other_shows:
            return self.redirect('/home/')

        # peel off the next one
        next_show_dir = other_shows[0]
        rest_of_show_dirs = other_shows[1:]

        # go to add the next show
        return self.redirect("/home/addShows/newShow?" + urlencode({'show_to_add': next_show_dir, 'other_shows': rest_of_show_dirs}, True))


class AddExistingShowsHandler(BaseHandler):
    @authenticated
    def post(self, *args, **kwargs):
        """
        Receives a dir list and add them. Adds the ones with given TVDB IDs first, then forwards
        along to the newShow page.
        """
        shows_to_add = self.get_arguments('shows_to_add')
        prompt_for_settings = self.get_argument('promptForSettings')

        # grab a list of other shows to add, if provided
        shows_to_add = [unquote_plus(x) for x in shows_to_add]

        prompt_for_settings = checkbox_to_value(prompt_for_settings)

        series_id_given = []
        dirs_only = []
        # separate all the ones with series id
        for cur_dir in shows_to_add:
            split_vals = cur_dir.split('|')
            if split_vals:
                if len(split_vals) > 2:
                    series_provider_id, show_dir, series_id, show_name = split_extra_show(cur_dir)
                    if all([show_dir, series_id, show_name]):
                        series_id_given.append((series_provider_id, show_dir, int(series_id), show_name))
                else:
                    dirs_only.append(cur_dir)
            else:
                dirs_only.append(cur_dir)

        # if they want me to prompt for settings then I will just carry on to the newShow page
        if prompt_for_settings and shows_to_add:
            return self.redirect("/home/addShows/newShow?" + urlencode({'show_to_add': shows_to_add[0], 'other_shows': shows_to_add[1:]}, True))

        # if they don't want me to prompt for settings then I can just add all the nfo shows now
        num_added = 0
        for cur_show in series_id_given:
            series_provider_id, show_dir, series_id, show_name = cur_show

            if series_provider_id is not None and series_id is not None:
                # add the show
                sickrage.app.show_queue.add_show(SeriesProviderID[series_provider_id],
                                                 series_id,
                                                 show_dir,
                                                 default_status=sickrage.app.config.general.status_default,
                                                 quality=sickrage.app.config.general.quality_default,
                                                 flatten_folders=sickrage.app.config.general.flatten_folders_default,
                                                 subtitles=sickrage.app.config.subtitles.default,
                                                 anime=sickrage.app.config.general.anime_default,
                                                 search_format=sickrage.app.config.general.search_format_default,
                                                 default_status_after=sickrage.app.config.general.status_default_after,
                                                 scene=sickrage.app.config.general.scene_default,
                                                 skip_downloaded=sickrage.app.config.general.skip_downloaded_default)
                num_added += 1

        if num_added:
            sickrage.app.alerts.message(_("Shows Added"),
                                        _("Automatically added ") + str(num_added) + _(" from their existing metadata files"))

        # if we're done then go home
        if not dirs_only:
            return self.redirect('/home/')

        # for the remaining shows we need to prompt for each one, so forward this on to the newShow page
        return self.redirect("/home/addShows/newShow?" + urlencode({'show_to_add': dirs_only[0], 'other_shows': dirs_only[1:]}, True))



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################


import datetime
import threading

import sickrage
from sickrage.core.common import Quality, Qualities, EpisodeStatus
from sickrage.core.queues.search import BacklogSearchTask
from sickrage.core.tv.show.helpers import find_show, get_show_list


class BacklogSearcher(object):
    def __init__(self, *args, **kwargs):
        self.name = "BACKLOG"
        self.lock = threading.Lock()
        self.cycleTime = 21 / 60 / 24
        self.running = False
        self.amPaused = False
        self.amWaiting = False
        self.forced = False

    def task(self, force=False):
        if self.running and not force:
            return

        try:
            self.running = True

            # set thread name
            threading.currentThread().setName(self.name)

            # set cycle time
            self.cycleTime = sickrage.app.config.general.backlog_searcher_freq / 60 / 24

            self.forced = force

            self.search_backlog()
        finally:
            self.running = False

    def am_running(self):
        sickrage.app.log.debug("amWaiting: " + str(self.amWaiting) + ", running: " + str(self.running))
        return (not self.amWaiting) and self.running

    def search_backlog(self, series_id=None, series_provider_id=None):
        self.amPaused = False

        show_list = [find_show(series_id, series_provider_id)] if series_id and series_provider_id else get_show_list()

        from_date = datetime.date.min

        if not series_id and self.forced:
            sickrage.app.log.info("Running limited backlog on missed episodes " + str(sickrage.app.config.general.backlog_days) + " day(s) old")
            from_date = datetime.date.today() - datetime.timedelta(days=sickrage.app.config.general.backlog_days)
        else:
            sickrage.app.log.info('Running full backlog search on missed episodes for all shows')

        # go through non air-by-date shows and see if they need any episodes
        for curShow in show_list:
            if curShow.paused:
                sickrage.app.log.debug("Skipping search for {} because the show is paused".format(curShow.name))
                continue

            wanted = self._get_wanted(curShow, from_date)
            if not wanted:
                sickrage.app.log.debug("Nothing needs to be downloaded for {}, skipping".format(curShow.name))
                continue

            for season, episode in wanted:
                if (curShow.series_id, season, episode) in sickrage.app.search_queue.SNATCH_HISTORY:
                    sickrage.app.search_queue.SNATCH_HISTORY.remove((curShow.series_id, season, episode))

                sickrage.app.search_queue.put(BacklogSearchTask(curShow.series_id, curShow.series_provider_id, season, episode))

            if from_date == datetime.date.min and not series_id:
                self._set_last_backlog_search(curShow, datetime.datetime.now())
                curShow.save()

    @staticmethod
    def _get_wanted(show, from_date):
        any_qualities, best_qualities = Quality.split_quality(show.quality)

        sickrage.app.log.debug("Seeing if we need anything that's older then today for {}".format(show.name))

        # check through the list of statuses to see if we want any
        wanted = []
        for episode_object in show.episodes:
            if not episode_object.season > 0 or not datetime.date.today() > episode_object.airdate > from_date:
                continue

            cur_status, cur_quality = Quality.split_composite_status(episode_object.status)

            # if we need a better one then say yes
            if cur_status not in {EpisodeStatus.WANTED, EpisodeStatus.DOWNLOADED, EpisodeStatus.SNATCHED, EpisodeStatus.SNATCHED_PROPER}:
                continue

            if cur_status != EpisodeStatus.WANTED:
                if best_qualities:
                    if cur_quality in best_qualities:
                        continue
                    elif cur_quality != Qualities.UNKNOWN and cur_quality > max(best_qualities):
                        continue
                elif any_qualities:
                    if cur_quality in any_qualities:
                        continue
                    elif cur_quality != Qualities.UNKNOWN and cur_quality > max(any_qualities):
                        continue

            # skip upgrading quality of downloaded episodes if enabled
            if cur_status == EpisodeStatus.DOWNLOADED and show.skip_downloaded:
                continue

            wanted += [(episode_object.season, episode_object.episode)]

        return wanted

    @staticmethod
    def _get_last_backlog_search(show):
        sickrage.app.log.debug("Retrieving the last check time from the DB")
        return show.last_backlog_search

    @staticmethod
    def _set_last_backlog_search(show, when):
        sickrage.app.log.debug("Setting the last backlog in the DB to {}".format(when))
        show.last_backlog_search = when



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################
from marshmallow_enum import EnumField
from marshmallow_sqlalchemy import SQLAlchemyAutoSchema

from sickrage.core.common import EpisodeStatus, Qualities
from sickrage.core.databases.main import MainDB
from sickrage.core.enums import SearchFormat, SeriesProviderID
from sickrage.core.helpers import camelcase


class TVShowSchema(SQLAlchemyAutoSchema):
    search_format = EnumField(SearchFormat)
    series_provider_id = EnumField(SeriesProviderID)
    default_ep_status = EnumField(EpisodeStatus)
    quality = EnumField(Qualities)

    class Meta:
        model = MainDB.TVShow
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class TVEpisodeSchema(SQLAlchemyAutoSchema):
    series_provider_id = EnumField(SeriesProviderID)
    status = EnumField(EpisodeStatus)

    class Meta:
        model = MainDB.TVEpisode
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class IMDbInfoSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = MainDB.IMDbInfo
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class BlacklistSchema(SQLAlchemyAutoSchema):
    series_provider_id = EnumField(SeriesProviderID)

    class Meta:
        model = MainDB.Blacklist
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class WhitelistSchema(SQLAlchemyAutoSchema):
    series_provider_id = EnumField(SeriesProviderID)

    class Meta:
        model = MainDB.Whitelist
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class HistoryScheme(SQLAlchemyAutoSchema):
    series_provider_id = EnumField(SeriesProviderID)
    quality = EnumField(Qualities)

    class Meta:
        model = MainDB.History
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class FailedSnatchHistoryScheme(SQLAlchemyAutoSchema):
    series_provider_id = EnumField(SeriesProviderID)
    old_status = EnumField(EpisodeStatus)

    class Meta:
        model = MainDB.FailedSnatchHistory
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: server_certificate_v1.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='server_certificate_v1.proto',
  package='app.protobufs.v1',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1bserver_certificate_v1.proto\x12\x10\x61pp.protobufs.v1\"J\n\x1eSavedServerCertificateResponse\x12\x13\n\x0b\x63\x65rtificate\x18\x01 \x01(\t\x12\x13\n\x0bprivate_key\x18\x02 \x01(\tb\x06proto3'
)




_SAVEDSERVERCERTIFICATERESPONSE = _descriptor.Descriptor(
  name='SavedServerCertificateResponse',
  full_name='app.protobufs.v1.SavedServerCertificateResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='certificate', full_name='app.protobufs.v1.SavedServerCertificateResponse.certificate', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='private_key', full_name='app.protobufs.v1.SavedServerCertificateResponse.private_key', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=49,
  serialized_end=123,
)

DESCRIPTOR.message_types_by_name['SavedServerCertificateResponse'] = _SAVEDSERVERCERTIFICATERESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

SavedServerCertificateResponse = _reflection.GeneratedProtocolMessageType('SavedServerCertificateResponse', (_message.Message,), {
  'DESCRIPTOR' : _SAVEDSERVERCERTIFICATERESPONSE,
  '__module__' : 'server_certificate_v1_pb2'
  # @@protoc_insertion_point(class_scope:app.protobufs.v1.SavedServerCertificateResponse)
  })
_sym_db.RegisterMessage(SavedServerCertificateResponse)


# @@protoc_insertion_point(module_scope)



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################


from tornado.web import authenticated

import sickrage
from sickrage.core.webserver import ConfigWebHandler
from sickrage.core.webserver.handlers.base import BaseHandler


class ConfigQualitySettingsHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        return self.render('config/quality_settings.mako',
                           submenu=ConfigWebHandler.menu,
                           title=_('Config - Quality Settings'),
                           header=_('Quality Settings'),
                           topmenu='config',
                           controller='config',
                           action='quality_settings')


class SaveQualitiesHandler(BaseHandler):
    @authenticated
    def post(self, *args, **kwargs):
        for quality in sickrage.app.config.quality_sizes.keys():
            quality_size_min = self.get_argument(f"{quality}_min")
            quality_size_max = self.get_argument(f"{quality}_max")
            sickrage.app.config.quality_sizes[quality].min_size = int(quality_size_min)
            sickrage.app.config.quality_sizes[quality].max_size = int(quality_size_max)

        sickrage.app.config.save()

        sickrage.app.alerts.message(_('[QUALITY SETTINGS] Configuration Saved to Database'))

        return self.redirect("/config/qualitySettings/")



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca/
# Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.


# CPU Presets for sleep timers
import enum
import operator
import pathlib
import re
from functools import reduce

from aenum import IntEnum, extend_enum

from sickrage.core.helpers.metadata import get_file_metadata, get_resolution

countryList = {'Australia': 'AU',
               'Canada': 'CA',
               'USA': 'US'}

dateFormat = '%Y-%m-%d'
dateTimeFormat = '%Y-%m-%d %H:%M:%S'
timeFormat = '%A %I:%M %p'

# Other constants
MULTI_EP_RESULT = -1
SEASON_RESULT = -2


class EpisodeStatus(IntEnum):
    UNKNOWN = -1  # SHOULD NEVER HAPPEN
    UNAIRED = 1  # EPISODES THAT HAVEN'T AIRED YET
    SNATCHED = 2  # QUALIFIED WITH QUALITY
    WANTED = 3  # EPISODES WE DON'T HAVE BUT WANT TO GET
    DOWNLOADED = 4  # QUALIFIED WITH QUALITY
    SKIPPED = 5  # EPISODES WE DON'T WANT
    ARCHIVED = 6  # EPISODES THAT YOU DON'T HAVE LOCALLY (COUNTS TOWARD DOWNLOAD COMPLETION STATS)
    IGNORED = 7  # EPISODES THAT YOU DON'T WANT INCLUDED IN YOUR DOWNLOAD STATS
    SNATCHED_PROPER = 9  # QUALIFIED WITH QUALITY
    SUBTITLED = 10  # QUALIFIED WITH QUALITY
    FAILED = 11  # EPISODE DOWNLOADED OR SNATCHED WE DON'T WANT
    SNATCHED_BEST = 12  # EPISODE REDOWNLOADED USING BEST QUALITY
    MISSED = 13

    @classmethod
    def _strings(cls):
        return {
            cls.UNKNOWN.name: "Unknown",
            cls.UNAIRED.name: "Unaired",
            cls.SNATCHED.name: "Snatched",
            cls.SNATCHED_PROPER.name: "Snatched (Proper)",
            cls.SNATCHED_BEST.name: "Snatched (Best)",
            cls.DOWNLOADED.name: "Downloaded",
            cls.SKIPPED.name: "Skipped",
            cls.WANTED.name: "Wanted",
            cls.ARCHIVED.name: "Archived",
            cls.IGNORED.name: "Ignored",
            cls.SUBTITLED.name: "Subtitled",
            cls.FAILED.name: "Failed",
            cls.MISSED.name: "Missed"
        }

    @classmethod
    def _prefix_strings(cls):
        return {
            cls.DOWNLOADED.name: _("Downloaded"),
            cls.SNATCHED.name: _("Snatched"),
            cls.SNATCHED_PROPER.name: _("Snatched (Proper)"),
            cls.SNATCHED_BEST.name: _("Snatched (Best)"),
            cls.ARCHIVED.name: _("Archived"),
            cls.FAILED.name: _("Failed"),
            cls.MISSED.name: _("Missed")
        }

    @property
    def display_name(self):
        status, quality = Quality.split_composite_status(self)
        if quality == Qualities.NONE:
            return self._strings()[status.name]
        return self._strings()[status.name] + " (" + quality.display_name + ")"

    @property
    def prefix_name(self):
        return self._prefix_strings()[self.name]

    @staticmethod
    def composites(status):
        return {
            EpisodeStatus.DOWNLOADED: [EpisodeStatus[f"{EpisodeStatus.DOWNLOADED.name}_{q.name}"] for q in Qualities if not q.is_preset],
            EpisodeStatus.SNATCHED: [EpisodeStatus[f"{EpisodeStatus.SNATCHED.name}_{q.name}"] for q in Qualities if not q.is_preset],
            EpisodeStatus.SNATCHED_PROPER: [EpisodeStatus[f"{EpisodeStatus.SNATCHED_PROPER.name}_{q.name}"] for q in Qualities if not q.is_preset],
            EpisodeStatus.SNATCHED_BEST: [EpisodeStatus[f"{EpisodeStatus.SNATCHED_BEST.name}_{q.name}"] for q in Qualities if not q.is_preset],
            EpisodeStatus.ARCHIVED: [EpisodeStatus[f"{EpisodeStatus.ARCHIVED.name}_{q.name}"] for q in Qualities if not q.is_preset],
            EpisodeStatus.FAILED: [EpisodeStatus[f"{EpisodeStatus.FAILED.name}_{q.name}"] for q in Qualities if not q.is_preset],
            EpisodeStatus.IGNORED: [EpisodeStatus[f"{EpisodeStatus.IGNORED.name}_{q.name}"] for q in Qualities if not q.is_preset],
        }[status]


class Overview(enum.Enum):
    UNAIRED = EpisodeStatus.UNAIRED.value  # 1
    SNATCHED = EpisodeStatus.SNATCHED.value  # 2
    WANTED = EpisodeStatus.WANTED.value  # 3
    GOOD = EpisodeStatus.DOWNLOADED.value  # 4
    SKIPPED = EpisodeStatus.SKIPPED.value  # 5
    SNATCHED_PROPER = EpisodeStatus.SNATCHED_PROPER.value  # 9
    SNATCHED_BEST = EpisodeStatus.SNATCHED_BEST.value  # 12
    MISSED = EpisodeStatus.MISSED.value  # 13
    LOW_QUALITY = 50

    @property
    def _strings(self):
        return {
            self.SKIPPED.name: "skipped",
            self.WANTED.name: "wanted",
            self.LOW_QUALITY.name: "low-quality",
            self.GOOD.name: "good",
            self.UNAIRED.name: "unaired",
            self.SNATCHED.name: "snatched",
            self.SNATCHED_BEST.name: "snatched",
            self.SNATCHED_PROPER.name: "snatched",
            self.MISSED.name: "missed"
        }

    @property
    def css_name(self):
        return self._strings[self.name]


class Quality(object):
    @staticmethod
    def combine_qualities(anyQualities, bestQualities):
        any_quality = 0
        best_quality = 0

        if anyQualities:
            any_quality = reduce(operator.or_, anyQualities, any_quality)
        if bestQualities:
            best_quality = reduce(operator.or_, bestQualities, best_quality)

        return any_quality | (best_quality << 16)

    @staticmethod
    def split_quality(quality):
        any_qualities = [quality_flag for quality_flag in Qualities if
                         quality_flag in Qualities(quality) and quality_flag and not quality_flag.is_preset]

        best_qualities = [quality_flag for quality_flag in Qualities if
                          quality_flag in Qualities(quality >> 16) and quality_flag and not quality_flag.is_preset]

        return sorted(any_qualities), sorted(best_qualities)

    @staticmethod
    def name_quality(name, anime=False):
        """
        Return The quality from an episode File renamed by SiCKRAGE
        If no quality is achieved it will try sceneQuality regex

        :param anime: Boolean to indicate if the show we're resolving is Anime
        :return: Quality prefix
        """

        # Try Scene names first
        quality = Quality.scene_quality(name, anime)
        if quality != Qualities.UNKNOWN:
            return quality

        quality = Quality.quality_from_file_meta(name)
        if quality != Qualities.UNKNOWN:
            return quality

        if name.lower().endswith(".ts"):
            return Qualities.RAWHDTV

        return Qualities.UNKNOWN

    @staticmethod
    def scene_quality(name, anime=False):
        """
        Return The quality from the scene episode File

        :param name: Episode filename to analyse
        :param anime: Boolean to indicate if the show we're resolving is Anime
        :return: Quality prefix
        """

        ret = Qualities.UNKNOWN
        if not name:
            return ret

        name = pathlib.Path(name).name

        check_name = lambda l, func: func([re.search(x, name, re.I) for x in l])

        if anime:
            dvd_options = check_name([r"dvd", r"dvdrip"], any)
            blue_ray_options = check_name([r"BD", r"blue?-?ray"], any)
            sd_options = check_name([r"360p", r"480p", r"848x480", r"XviD"], any)
            hd_options = check_name([r"720p", r"1280x720", r"960x720"], any)
            full_hd = check_name([r"1080p", r"1920x1080"], any)

            if sd_options and not blue_ray_options and not dvd_options:
                ret = Qualities.SDTV
            elif dvd_options:
                ret = Qualities.SDDVD
            elif hd_options and not blue_ray_options and not full_hd:
                ret = Qualities.HDTV
            elif full_hd and not blue_ray_options and not hd_options:
                ret = Qualities.FULLHDTV
            elif hd_options and not blue_ray_options and not full_hd:
                ret = Qualities.HDWEBDL
            elif blue_ray_options and hd_options and not full_hd:
                ret = Qualities.HDBLURAY
            elif blue_ray_options and full_hd and not hd_options:
                ret = Qualities.FULLHDBLURAY

            return ret

        if (check_name([r"480p|\bweb\b|web.?dl|web(rip|mux|hd)|[sph]d.?tv|dsr|tv(rip|mux)|satrip", r"xvid|divx|[xh].?26[45]"], all)
                and not check_name([r"(720|1080|2160|4320)[pi]"], all)
                and not check_name([r"hr.ws.pdtv.[xh].?26[45]"], any)):
            ret = Qualities.SDTV
        elif (check_name([r"dvd(rip|mux)|b[rd](rip|mux)|blue?-?ray", r"xvid|divx|[xh].?26[45]"], all)
              and not check_name([r"(720|1080|2160|4320)[pi]"], all)
              and not check_name([r"hr.ws.pdtv.[xh].?26[45]"], any)):
            ret = Qualities.SDDVD
        elif (check_name([r"720p", r"hd.?tv", r"[xh].?26[45]"], all)
              or check_name([r"720p", r"hevc", r"[xh].?26[45]"], all)
              or check_name([r"hr.ws.pdtv.[xh].?26[45]"], any) and not check_name([r"1080[pi]"], all)):
            ret = Qualities.HDTV
        elif (check_name([r"720p|1080i", r"hd.?tv", r"mpeg-?2"], all)
              or check_name([r"1080[pi].hdtv", r"h.?26[45]"], all)):
            ret = Qualities.RAWHDTV
        elif (check_name([r"1080p", r"hd.?tv", r"[xh].?26[45]"], all)
              or check_name([r"1080p", r"hevc", r"[xh].?26[45]"], all)):
            ret = Qualities.FULLHDTV
        elif (check_name([r"720p", r"\bweb\b|web.?dl|web(rip|mux|hd)"], all)
              or check_name([r"720p", r"itunes", r"[xh].?26[45]"], all)):
            ret = Qualities.HDWEBDL
        elif (check_name([r"1080p", r"\bweb\b|web.?dl|web(rip|mux|hd)"], all)
              or check_name([r"1080p", r"itunes", r"[xh].?26[45]"], all)):
            ret = Qualities.FULLHDWEBDL
        elif check_name([r"720p", r"blue?-?ray|hddvd|b[rd](rip|mux)", r"[xh].?26[45]"], all):
            ret = Qualities.HDBLURAY
        elif check_name([r"1080p", r"blue?-?ray|hddvd|b[rd](rip|mux)", r"[xh].?26[45]"], all):
            ret = Qualities.FULLHDBLURAY
        elif check_name([r"2160p", r"hd.?tv", r"[xh].?26[45]"], all):
            ret = Qualities.UHD_4K_TV
        elif check_name([r"4320p", r"hd.?tv", r"[xh].?26[45]"], all):
            ret = Qualities.UHD_8K_TV
        elif (check_name([r"2160p", r"\bweb\b|web.?dl|web(rip|mux|hd)"], all)
              or check_name([r"2160p", r"itunes", r"[xh].?26[45]"], all)):
            ret = Qualities.UHD_4K_WEBDL
        elif (check_name([r"4320p", r"\bweb\b|web.?dl|web(rip|mux|hd)"], all)
              or check_name([r"4320p", r"itunes", r"[xh].?26[45]"], all)):
            ret = Qualities.UHD_8K_WEBDL
        elif check_name([r"2160p", r"blue?-?ray|hddvd|b[rd](rip|mux)", r"[xh].?26[45]"], all):
            ret = Qualities.UHD_4K_BLURAY
        elif check_name([r"4320p", r"blue?-?ray|hddvd|b[rd](rip|mux)", r"[xh].?26[45]"], all):
            ret = Qualities.UHD_8K_BLURAY

        return ret

    @staticmethod
    def composite_status(status, quality):
        return EpisodeStatus(status + 100 * quality)

    @staticmethod
    def split_composite_status(status):
        """Returns a tuple containing (status, quality)"""
        if status == EpisodeStatus.UNKNOWN:
            return status, Qualities.UNKNOWN

        for q in sorted(Qualities, reverse=True):
            if status > q * 100:
                return EpisodeStatus(status - q * 100), q

        return status, Qualities.NONE

    @staticmethod
    def quality_from_file_meta(filename):
        """
        Get quality from file metadata

        :param filename: Filename to analyse
        :return: Quality prefix
        """

        data = {}
        quality = Qualities.UNKNOWN

        try:
            if pathlib.Path(filename).is_file():
                meta = get_file_metadata(filename)

                if meta.get('resolution_width') and meta.get('resolution_height'):
                    data['resolution_width'] = meta.get('resolution_width')
                    data['resolution_height'] = meta.get('resolution_height')
                    data['aspect'] = round(float(meta.get('resolution_width')) / meta.get('resolution_height', 1), 2)
                else:
                    data.update(get_resolution(filename))

                base_filename = pathlib.Path(filename).name
                bluray = re.search(r"blue?-?ray|hddvd|b[rd](rip|mux)", base_filename, re.I) is not None
                webdl = re.search(r"\bweb\b|web.?dl|web(rip|mux|hd)", base_filename, re.I) is not None

                if 3240 < data['resolution_height']:
                    quality = ((Qualities.UHD_8K_TV, Qualities.UHD_8K_BLURAY)[bluray], Qualities.UHD_8K_WEBDL)[webdl]
                if 1620 < data['resolution_height'] <= 3240:
                    quality = ((Qualities.UHD_4K_TV, Qualities.UHD_4K_BLURAY)[bluray], Qualities.UHD_4K_WEBDL)[webdl]
                elif 800 < data['resolution_height'] <= 1620:
                    quality = ((Qualities.FULLHDTV, Qualities.FULLHDBLURAY)[bluray], Qualities.FULLHDWEBDL)[webdl]
                elif 680 < data['resolution_height'] < 800:
                    quality = ((Qualities.HDTV, Qualities.HDBLURAY)[bluray], Qualities.HDWEBDL)[webdl]
                elif data['resolution_height'] < 680:
                    quality = (Qualities.SDTV, Qualities.SDDVD)[re.search(r'dvd|b[rd]rip|blue?-?ray', base_filename, re.I) is not None]
        except Exception:
            pass

        return quality

    @staticmethod
    def scene_quality_from_name(name, quality):
        """
        Get scene naming parameters from filename and quality

        :param name: Filename to check
        :param quality: int of quality to make sure we get the right rip type
        :return: encoder type for scene quality naming
        """
        codecList = ['xvid', 'divx']
        x264List = ['x264', 'x 264', 'x.264']
        h264List = ['h264', 'h 264', 'h.264', 'avc']
        x265List = ['x265', 'x 265', 'x.265']
        h265List = ['h265', 'h 265', 'h.265', 'hevc']
        codecList.extend(x264List + h264List + x265List + h265List)

        found_codecs = {}
        found_codec = None
        rip_type = ""

        for codec in codecList:
            if codec in name.lower():
                found_codecs[name.lower().rfind(codec)] = codec

        if found_codecs:
            sorted_codecs = sorted(found_codecs, reverse=True)
            found_codec = found_codecs[list(sorted_codecs)[0]]

        if quality == Qualities.SDDVD:
            if re.search(r"b(r|d|rd)?([- .])?(rip|mux)", name.lower()):
                rip_type = " BDRip"
            elif re.search(r"(dvd)([- .])?(rip|mux)?", name.lower()):
                rip_type = " DVDRip"
            else:
                rip_type = ""

        if found_codec:
            if codecList[0] in found_codec:
                found_codec = 'XviD'
            elif codecList[1] in found_codec:
                found_codec = 'DivX'
            elif found_codec in x264List:
                found_codec = x264List[0]
            elif found_codec in h264List:
                found_codec = h264List[0]
            elif found_codec in x265List:
                found_codec = x265List[0]
            elif found_codec in h265List:
                found_codec = h265List[0]

            if quality == Qualities.SDDVD:
                return rip_type + " " + found_codec
            else:
                return " " + found_codec
        elif quality == Qualities.SDDVD:
            return rip_type
        else:
            return ""

    @staticmethod
    def status_from_name(name, assume=True, anime=False):
        """
        Get a status object from filename

        :param name: Filename to check
        :param assume: boolean to assume quality by extension if we can't figure it out
        :param anime: boolean to enable anime parsing
        :return: Composite status/quality object
        """
        quality = Quality.name_quality(name, anime)
        return Quality.composite_status(EpisodeStatus.DOWNLOADED, quality)

    @staticmethod
    def from_guessit(guess):
        """
        Return a Quality from a guessit dict.
        :param guess: guessit dict
        :type guess: dict
        :return: quality
        :rtype: int
        """
        guessit_map = {
            '720p': {
                'HDTV': Qualities.HDTV,
                'Web': Qualities.HDWEBDL,
                'Blu-ray': Qualities.HDBLURAY,
            },
            '1080i': Qualities.RAWHDTV,
            '1080p': {
                'HDTV': Qualities.FULLHDTV,
                'Web': Qualities.FULLHDWEBDL,
                'Blu-ray': Qualities.FULLHDBLURAY
            },
            '2160p': {
                'HDTV': Qualities.UHD_4K_TV,
                'Web': Qualities.UHD_4K_WEBDL,
                'Blu-ray': Qualities.UHD_4K_BLURAY
            },
            '4320p': {
                'HDTV': Qualities.UHD_8K_TV,
                'Web': Qualities.UHD_8K_WEBDL,
                'Blu-ray': Qualities.UHD_8K_BLURAY
            }
        }

        screen_size = guess.get('screen_size')
        source = guess.get('source')

        if not screen_size or isinstance(screen_size, list):
            return Qualities.UNKNOWN

        source_map = guessit_map.get(screen_size)
        if not source_map:
            return Qualities.UNKNOWN

        if isinstance(source_map, int):
            return source_map

        if not source or isinstance(source, list):
            return Qualities.UNKNOWN

        quality = source_map.get(source)
        return quality if quality is not None else Qualities.UNKNOWN

    @staticmethod
    def to_guessit(quality):
        """
        Return a guessit dict containing 'screen_size and source' from a Quality.
        :param quality: a quality
        :type quality: int
        :return: dict {'screen_size': <screen_size>, 'source': <source>}
        :rtype: dict (str, str)
        """
        if quality not in Qualities:
            quality = Qualities.UNKNOWN

        screen_size = Quality.to_guessit_screen_size(quality)
        source = Quality.to_guessit_source(quality)

        result = {}
        if screen_size:
            result['screen_size'] = screen_size
        if source:
            result['source'] = source

        return result

    @staticmethod
    def to_guessit_source(quality):
        """
        Return a guessit source from a Quality.
        :param quality: the quality
        :type quality: int
        :return: guessit source
        :rtype: str
        """

        source_map = {
            Qualities.ANYHDTV | Qualities.UHD_4K_TV | Qualities.UHD_8K_TV: 'HDTV',
            Qualities.ANYWEBDL | Qualities.UHD_4K_WEBDL | Qualities.UHD_8K_WEBDL: 'Web',
            Qualities.ANYBLURAY | Qualities.UHD_4K_BLURAY | Qualities.UHD_8K_BLURAY: 'Blu-ray'
        }

        for quality_set, source in source_map.items():
            if quality_set & quality:
                return source

    @staticmethod
    def to_guessit_screen_size(quality):
        """
        Return a guessit screen_size from a Quality.
        :param quality: the quality
        :type quality: int
        :return: guessit screen_size
        :rtype: str
        """

        screen_size_map = {
            Qualities.HDTV | Qualities.HDWEBDL | Qualities.HDBLURAY: '720p',
            Qualities.RAWHDTV: '1080i',
            Qualities.FULLHDTV | Qualities.FULLHDWEBDL | Qualities.FULLHDBLURAY: '1080p',
            Qualities.UHD_4K_TV | Qualities.UHD_4K_WEBDL | Qualities.UHD_4K_BLURAY: '2160p',
            Qualities.UHD_8K_TV | Qualities.UHD_8K_WEBDL | Qualities.UHD_8K_BLURAY: '4320p',
        }

        for quality_set, screen_size in screen_size_map.items():
            if quality_set & quality:
                return screen_size


class Qualities(enum.IntFlag):
    NONE = 0  # 0
    SDTV = 1  # 1
    SDDVD = 1 << 1  # 2
    HDTV = 1 << 2  # 4
    RAWHDTV = 1 << 3  # 8  -- 720P/1080I MPEG2 (TROLLHD RELEASES)
    FULLHDTV = 1 << 4  # 16 -- 1080P HDTV (QCF RELEASES)
    HDWEBDL = 1 << 5  # 32
    FULLHDWEBDL = 1 << 6  # 64 -- 1080P WEB-DL
    HDBLURAY = 1 << 7  # 128
    FULLHDBLURAY = 1 << 8  # 256
    UHD_4K_TV = 1 << 9  # 512 -- 2160P AKA 4K UHD AKA UHD-1
    UHD_4K_WEBDL = 1 << 10  # 1024
    UHD_4K_BLURAY = 1 << 11  # 2048
    UHD_8K_TV = 1 << 12  # 4096 -- 4320P AKA 8K UHD AKA UHD-2
    UHD_8K_WEBDL = 1 << 13  # 8192
    UHD_8K_BLURAY = 1 << 14  # 16384

    ANYHDTV = HDTV | FULLHDTV  # 20
    ANYWEBDL = HDWEBDL | FULLHDWEBDL  # 96
    ANYBLURAY = HDBLURAY | FULLHDBLURAY

    UNKNOWN = 1 << 15  # 32768

    # Presets
    SD = Quality.combine_qualities([SDTV, SDDVD], [])
    HD720P = Quality.combine_qualities([HDTV, HDWEBDL, HDBLURAY], [])
    HD1080P = Quality.combine_qualities([FULLHDTV, FULLHDWEBDL, FULLHDBLURAY], [])
    HD = Quality.combine_qualities([HD720P, HD1080P, RAWHDTV], [])
    UHD_4K = Quality.combine_qualities([UHD_4K_TV, UHD_4K_WEBDL, UHD_4K_BLURAY], [])
    UHD_8K = Quality.combine_qualities([UHD_8K_TV, UHD_8K_WEBDL, UHD_8K_BLURAY], [])
    UHD = Quality.combine_qualities([UHD_4K, UHD_8K], [])
    ANY = Quality.combine_qualities([SD, HD, UHD], [])
    ANY_PLUS_UNKNOWN = Quality.combine_qualities([UNKNOWN, SD, HD, UHD], [])

    @property
    def _strings(self):
        return {
            self.NONE.name: "N/A",
            self.UNKNOWN.name: "Unknown",
            self.SDTV.name: "SDTV",
            self.SDDVD.name: "SD DVD",
            self.HDTV.name: "720p HDTV",
            self.RAWHDTV.name: "RawHD",
            self.FULLHDTV.name: "1080p HDTV",
            self.HDWEBDL.name: "720p WEB-DL",
            self.FULLHDWEBDL.name: "1080p WEB-DL",
            self.HDBLURAY.name: "720p BluRay",
            self.FULLHDBLURAY.name: "1080p BluRay",
            self.UHD_4K_TV.name: "4K UHD TV",
            self.UHD_8K_TV.name: "8K UHD TV",
            self.UHD_4K_WEBDL.name: "4K UHD WEB-DL",
            self.UHD_8K_WEBDL.name: "8K UHD WEB-DL",
            self.UHD_4K_BLURAY.name: "4K UHD BluRay",
            self.UHD_8K_BLURAY.name: "8K UHD BluRay"
        }

    @property
    def _preset_strings(self):
        return {
            self.SD.name: "SD",
            self.HD.name: "HD",
            self.HD720P.name: "HD720p",
            self.HD1080P.name: "HD1080p",
            self.UHD.name: "UHD",
            self.UHD_4K.name: "UHD-4K",
            self.UHD_8K.name: "UHD-8K",
            self.ANY.name: "Any",
            self.ANY_PLUS_UNKNOWN.name: "Any + Unknown"
        }

    @property
    def _scene_strings(self):
        return {
            self.NONE.name: "N/A",
            self.UNKNOWN.name: "Unknown",
            self.SDTV.name: "HDTV",
            self.SDDVD.name: "",
            self.HDTV.name: "720p HDTV",
            self.RAWHDTV.name: "1080i HDTV",
            self.FULLHDTV.name: "1080p HDTV",
            self.HDWEBDL.name: "720p WEB-DL",
            self.FULLHDWEBDL.name: "1080p WEB-DL",
            self.HDBLURAY.name: "720p BluRay",
            self.FULLHDBLURAY.name: "1080p BluRay",
            self.UHD_4K_TV.name: "4K UHD TV",
            self.UHD_8K_TV.name: "8K UHD TV",
            self.UHD_4K_WEBDL.name: "4K UHD WEB-DL",
            self.UHD_8K_WEBDL.name: "8K UHD WEB-DL",
            self.UHD_4K_BLURAY.name: "4K UHD BluRay",
            self.UHD_8K_BLURAY.name: "8K UHD BluRay"
        }

    @property
    def _css_strings(self):
        return {
            self.NONE.name: "N/A",
            self.UNKNOWN.name: "Unknown",
            self.SDTV.name: "SDTV",
            self.SDDVD.name: "SDDVD",
            self.HDTV.name: "HD720p",
            self.RAWHDTV.name: "RawHD",
            self.FULLHDTV.name: "HD1080p",
            self.HDWEBDL.name: "HD720p",
            self.FULLHDWEBDL.name: "HD1080p",
            self.HDBLURAY.name: "HD720p",
            self.FULLHDBLURAY.name: "HD1080p",
            self.UHD_4K_TV.name: "UHD-4K",
            self.UHD_8K_TV.name: "UHD-8K",
            self.UHD_4K_WEBDL.name: "UHD-4K",
            self.UHD_8K_WEBDL.name: "UHD-8K",
            self.UHD_4K_BLURAY.name: "UHD-4K",
            self.UHD_8K_BLURAY.name: "UHD-8K",
            self.ANYHDTV.name: "any-hd",
            self.ANYWEBDL.name: "any-hd",
            self.ANYBLURAY.name: "any-hd"
        }

    @property
    def _combined_strings(self):
        return {
            self.ANYHDTV.name: "HDTV",
            self.ANYWEBDL.name: "WEB-DL",
            self.ANYBLURAY.name: "BluRay"
        }

    @property
    def display_name(self):
        if self.name in self._strings:
            return self._strings[self.name]
        elif self.name in self._preset_strings:
            return self._preset_strings[self.name]
        elif self.name in self._combined_strings:
            return self._combined_strings[self.name]
        return "Custom"

    @property
    def scene_name(self):
        if self.name in self._scene_strings:
            return self._scene_strings[self.name]
        return ""

    @property
    def css_name(self):
        if self.name in self._css_strings:
            return self._css_strings[self.name]
        elif self.name in self._preset_strings:
            return self._preset_strings[self.name]
        return ""

    @property
    def is_preset(self):
        return self.name in self._preset_strings

    @property
    def is_combined(self):
        return self.name in self._combined_strings


# extend episode status enum class with composite statuses
[extend_enum(EpisodeStatus, f"{status.name}_{q.name}", status + 100 * q)
 for status in list(EpisodeStatus).copy() for q in Qualities if not q.is_preset and status in [EpisodeStatus.DOWNLOADED,
                                                                                               EpisodeStatus.SNATCHED,
                                                                                               EpisodeStatus.SNATCHED_BEST,
                                                                                               EpisodeStatus.SNATCHED_PROPER,
                                                                                               EpisodeStatus.ARCHIVED,
                                                                                               EpisodeStatus.FAILED,
                                                                                               EpisodeStatus.IGNORED,
                                                                                               EpisodeStatus.SUBTITLED]]



# Author: Frank Fenton
# URL: https://sickrage.ca
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.


import datetime
import os
import threading
import traceback

import sickrage
from sickrage.core.common import EpisodeStatus
from sickrage.core.enums import TraktAddMethod, SeriesProviderID
from sickrage.core.exceptions import EpisodeNotFoundException
from sickrage.core.helpers import sanitize_file_name, make_dir, chmod_as_parent, flatten
from sickrage.core.queues.search import BacklogSearchTask
from sickrage.core.traktapi import TraktAPI
from sickrage.core.tv.show.helpers import find_show, get_show_list


def set_episode_to_wanted(show, s, e):
    """
    Sets an episode to wanted, only if it is currently skipped
    """
    try:
        epObj = show.get_episode(s, e)
        if epObj.status != EpisodeStatus.SKIPPED or not epObj.airdate > datetime.date.min:
            return

        sickrage.app.log.info("Setting episode %s S%02dE%02d to wanted" % (show.name, s, e))

        epObj.status = EpisodeStatus.WANTED
        epObj.save()

        sickrage.app.search_queue.put(BacklogSearchTask(show.series_id, show.series_provider_id, epObj.season, epObj.episode))

        sickrage.app.log.info("Starting backlog search for %s S%02dE%02d because some episodes were set to wanted" % (show.name, s, e))
    except EpisodeNotFoundException as e:
        pass


class TraktSearcher(object):
    def __init__(self):
        self.name = "TRAKTSEARCHER"

        self.todoBacklog = []
        self.todoWanted = []
        self.ShowWatchlist = {}
        self.EpisodeWatchlist = {}
        self.Collectionlist = {}
        self.running = False

    def task(self, force=False):
        if self.running or not sickrage.app.config.trakt.enable and not force:
            return

        try:
            self.running = True

            # set thread name
            threading.currentThread().setName(self.name)

            self.todoWanted = []  # its about to all get re-added
            if len(sickrage.app.config.general.root_dirs.split('|')) < 2:
                sickrage.app.log.warning("No default root directory")
                return

            # add shows from tv watchlist
            if sickrage.app.config.trakt.sync_watchlist:
                try:
                    self.sync_watchlist()
                except Exception:
                    sickrage.app.log.debug(traceback.format_exc())

            # add shows from tv collection
            if sickrage.app.config.trakt.sync:
                try:
                    self.sync_collection()
                except Exception:
                    sickrage.app.log.debug(traceback.format_exc())
        finally:
            self.running = False

    def sync_watchlist(self):
        sickrage.app.log.debug("Syncing SiCKRAGE with Trakt Watchlist")

        self.remove_show_from_sickrage()

        if self._get_show_watchlist():
            self.add_show_to_trakt_watch_list()
            self.update_shows()

        if self._get_episode_watchlist():
            self.add_episodes_to_trakt_watch_list()
            if sickrage.app.config.trakt.remove_show_from_sickrage:
                self.remove_episodes_from_trakt_watch_list()
            self.update_episodes()

    def sync_collection(self):
        sickrage.app.log.debug("Syncing SiCKRAGE with Trakt Collection")
        if self._get_show_collection():
            self.add_episodes_to_trakt_collection()
            if sickrage.app.config.trakt.sync_remove:
                self.remove_episodes_from_trakt_collection()

    def find_show_match(self, series_provider_id, series_id):
        trakt_show = None

        try:
            library = TraktAPI()["sync/collection"].shows() or {}
            if not library:
                sickrage.app.log.debug("No shows found in your library, aborting library update")
                return

            trakt_show = [x for __, x in library.items() if int(series_id) == int(x.ids[sickrage.app.series_providers[series_provider_id].trakt_id])]
        except Exception as e:
            sickrage.app.log.warning(f"Could not connect to Trakt service. Aborting library check. Error: {repr(e)}")

        return trakt_show

    def remove_show_from_trakt_library(self, show_obj):
        if self.find_show_match(show_obj.series_provider_id, show_obj.series_id):
            # URL parameters
            data = {
                'shows': [
                    {
                        'title': show_obj.name,
                        'year': show_obj.startyear,
                        'ids': {show_obj.series_provider.trakt_id: show_obj.series_id}
                    }
                ]
            }

            sickrage.app.log.debug(f"Removing {show_obj.name} from tv library")

            try:
                TraktAPI()["sync/collection"].remove(data)
            except Exception as e:
                sickrage.app.log.warning(f"Could not connect to Trakt service. Aborting removing show {show_obj.name} from Trakt library. Error: {repr(e)}")

    def add_show_to_trakt_library(self, show_obj):
        """
        Sends a request to trakt indicating that the given show and all its episodes is part of our library.

        show_obj: The TVShow object to add to trakt
        """
        data = {}

        if not self.find_show_match(show_obj.series_provider_id, show_obj.series_id):
            # URL parameters
            data = {
                'shows': [
                    {
                        'title': show_obj.name,
                        'year': show_obj.startyear,
                        'ids': {show_obj.series_provider.trakt_id: show_obj.series_id}
                    }
                ]
            }

        if len(data):
            sickrage.app.log.debug(f"Adding {show_obj.name} to tv library")

            try:
                TraktAPI()["sync/collection"].add(data)
            except Exception as e:
                sickrage.app.log.warning(f"Could not connect to Trakt service. Aborting adding show {show_obj.name} to Trakt library. Error: {repr(e)}")
                return

    def add_episodes_to_trakt_collection(self):
        trakt_data = []

        sickrage.app.log.debug("COLLECTION::SYNC::START - Look for Episodes to Add to Trakt Collection")

        for s in get_show_list():
            for e in s.episodes:
                trakt_id = s.series_provider.trakt_id
                if not self._check_in_list(trakt_id, str(e.series_id), e.season, e.episode, 'Collection'):
                    sickrage.app.log.debug(f"Adding Episode {s.name} S{e.season:02d}E{e.episode:02d} to collection")
                    trakt_data.append((e.series_id, s.series_provider_id, s.name, s.startyear, e.season, e.episode))

        if len(trakt_data):
            try:
                TraktAPI()["sync/collection"].add(self.trakt_bulk_data_generate(trakt_data))
                self._get_show_collection()
            except Exception as e:
                sickrage.app.log.warning(f"Could not connect to Trakt service. Error: {e}")

        sickrage.app.log.debug("COLLECTION::ADD::FINISH - Look for Episodes to Add to Trakt Collection")

    def remove_episodes_from_trakt_collection(self):
        trakt_data = []

        sickrage.app.log.debug(
            "COLLECTION::REMOVE::START - Look for Episodes to Remove From Trakt Collection")

        for s in get_show_list():
            for e in s.episodes:
                if e.location:
                    continue

                series_provider_trakt_id = s.series_provider.trakt_id
                if self._check_in_list(series_provider_trakt_id, str(e.series_id), e.season, e.episode, 'Collection'):
                    sickrage.app.log.debug(f"Removing Episode {s.name} S{e.season:02d}E{e.episode:02d} from collection")
                    trakt_data.append((e.series_id, s.series_provider_id, s.name, s.startyear, e.season, e.episode))

        if len(trakt_data):
            try:
                TraktAPI()["sync/collection"].remove(self.trakt_bulk_data_generate(trakt_data))
                self._get_show_collection()
            except Exception as e:
                sickrage.app.log.warning(f"Could not connect to Trakt service. Error: {e}")

        sickrage.app.log.debug(
            "COLLECTION::REMOVE::FINISH - Look for Episodes to Remove From Trakt Collection")

    def remove_episodes_from_trakt_watch_list(self):
        trakt_data = []

        sickrage.app.log.debug("WATCHLIST::REMOVE::START - Look for Episodes to Remove from Trakt Watchlist")

        for s in get_show_list():
            for e in s.episodes:
                series_provider_trakt_id = s.series_provider.trakt_id
                if self._check_in_list(series_provider_trakt_id, str(e.series_id), e.season, e.episode):
                    sickrage.app.log.debug(f"Removing Episode {s.name} S{e.season:02d}E{e.episode:02d} from watchlist")
                    trakt_data.append((e.series_id, s.series_provider_id, s.name, s.startyear, e.season, e.episode))

        if len(trakt_data):
            try:
                data = self.trakt_bulk_data_generate(trakt_data)
                TraktAPI()["sync/watchlist"].remove(data)
                self._get_episode_watchlist()
            except Exception as e:
                sickrage.app.log.warning(f"Could not connect to Trakt service. Error: {e}")

        sickrage.app.log.debug(
            "WATCHLIST::REMOVE::FINISH - Look for Episodes to Remove from Trakt Watchlist")

    def add_episodes_to_trakt_watch_list(self):
        trakt_data = []

        sickrage.app.log.debug("WATCHLIST::ADD::START - Look for Episodes to Add to Trakt Watchlist")

        for show_object in get_show_list():
            for episode_object in show_object.episodes:
                if episode_object.status in flatten(
                        [EpisodeStatus.composites(EpisodeStatus.SNATCHED), EpisodeStatus.composites(EpisodeStatus.SNATCHED_PROPER), EpisodeStatus.UNKNOWN,
                         EpisodeStatus.WANTED]):
                    continue

                trakt_id = show_object.series_provider.trakt_id
                if self._check_in_list(trakt_id, str(show_object.series_id), episode_object.season, episode_object.episode):
                    sickrage.app.log.debug(f"Adding Episode {show_object.name} S{episode_object.season:02d}E{episode_object.episode:02d} to watchlist")
                    trakt_data.append(
                        (show_object.series_id, show_object.series_provider_id, show_object.name, show_object.startyear, episode_object.season,
                         episode_object.episode))

        if len(trakt_data):
            try:
                data = self.trakt_bulk_data_generate(trakt_data)
                TraktAPI()["sync/watchlist"].add(data)
                self._get_episode_watchlist()
            except Exception as e:
                sickrage.app.log.warning(f"Could not connect to Trakt service. Error {e}")

        sickrage.app.log.debug("WATCHLIST::ADD::FINISH - Look for Episodes to Add to Trakt Watchlist")

    def add_show_to_trakt_watch_list(self):
        trakt_data = []

        sickrage.app.log.debug("SHOW_WATCHLIST::ADD::START - Look for Shows to Add to Trakt Watchlist")

        for show in get_show_list():
            series_provider_trakt_id = show.series_provider.trakt_id

            if not self._check_in_list(series_provider_trakt_id, str(show.series_id), 0, 0, 'Show'):
                sickrage.app.log.debug(f"Adding Show: Series Provider {series_provider_trakt_id} {str(show.series_id)} - {show.name} to Watchlist")

                show_el = {'title': show.name,
                           'year': show.startyear,
                           'ids': {series_provider_trakt_id: show.series_id}}

                trakt_data.append(show_el)

        if len(trakt_data):
            try:
                data = {'shows': trakt_data}
                TraktAPI()["sync/watchlist"].add(data)
                self._get_show_watchlist()
            except Exception as e:
                sickrage.app.log.warning(f"Could not connect to Trakt service. Error: {e}")

        sickrage.app.log.debug("SHOW_WATCHLIST::ADD::FINISH - Look for Shows to Add to Trakt Watchlist")

    def remove_show_from_sickrage(self):
        sickrage.app.log.debug("SHOW_SICKRAGE::REMOVE::START - Look for Shows to remove from SiCKRAGE")

        for show in get_show_list():
            if show.status == "Ended":
                try:
                    progress = TraktAPI()["shows"].get(show.imdb_id)
                except Exception as e:
                    sickrage.app.log.warning(
                        "Could not connect to Trakt service. Aborting removing show %s from SiCKRAGE. Error: %s" % (
                            show.name, repr(e)))
                    return

                if progress.status in ['canceled', 'ended']:
                    sickrage.app.show_queue.remove_show(show.series_id, show.series_provider_id, full=True)
                    sickrage.app.log.debug("Show: %s has been removed from SiCKRAGE" % show.name)

        sickrage.app.log.debug("SHOW_SICKRAGE::REMOVE::FINISH - Trakt Show Watchlist")

    def update_shows(self):
        sickrage.app.log.debug("SHOW_WATCHLIST::CHECK::START - Trakt Show Watchlist")

        if not len(self.ShowWatchlist):
            sickrage.app.log.debug("No shows found in your watchlist, aborting watchlist update")
            return

        for key, show in self.ShowWatchlist.items():
            # get traktID and series_provider_id values
            trakt_id, series_id = key

            series_provider_id = None
            for item in SeriesProviderID:
                if item.trakt_id == trakt_id:
                    series_provider_id = item

            if series_provider_id:
                if sickrage.app.config.trakt.method_add != TraktAddMethod.WHOLE_SHOW:
                    self.add_default_show(series_provider_id, series_id, show.title, EpisodeStatus.SKIPPED)
                else:
                    self.add_default_show(series_provider_id, series_id, show.title, EpisodeStatus.WANTED)

                if sickrage.app.config.trakt.method_add == TraktAddMethod.DOWNLOAD_PILOT_ONLY:
                    newShow = find_show(series_id, series_provider_id)

                    if newShow is not None:
                        set_episode_to_wanted(newShow, 1, 1)
                    else:
                        self.todoWanted.append((series_id, 1, 1))

        sickrage.app.log.debug("SHOW_WATCHLIST::CHECK::FINISH - Trakt Show Watchlist")

    def update_episodes(self):
        """
        Sets episodes to wanted that are in trakt watchlist
        """
        sickrage.app.log.debug("SHOW_WATCHLIST::CHECK::START - Trakt Episode Watchlist")

        if not len(self.EpisodeWatchlist):
            sickrage.app.log.debug("No episode found in your watchlist, aborting episode update")
            return

        managed_show = []

        for key, show in self.EpisodeWatchlist.items():
            # get traktID and series_provider_id values
            trakt_id, series_id = key

            series_provider_id = None
            for item in SeriesProviderID:
                if item.trakt_id == trakt_id:
                    series_provider_id = item

            if series_provider_id:
                show_object = find_show(series_id, series_provider_id)

                try:
                    if show_object is None:
                        if series_id not in managed_show:
                            self.add_default_show(series_provider_id, series_id, show.title, EpisodeStatus.SKIPPED)
                            managed_show.append(series_id)

                            for season_number, season in show.seasons.items():
                                for episode_number, _ in season.episodes.items():
                                    self.todoWanted.append((int(series_id), int(season_number), int(episode_number)))
                    else:
                        if show_object.series_provider_id == series_provider_id:
                            for season_number, season in show.seasons.items():
                                for episode_number, _ in season.episodes.items():
                                    set_episode_to_wanted(show_object, int(season_number), int(episode_number))
                except TypeError:
                    sickrage.app.log.debug(f"Could not parse the output from trakt for {show.title} ")

        sickrage.app.log.debug("SHOW_WATCHLIST::CHECK::FINISH - Trakt Episode Watchlist")

    @staticmethod
    def add_default_show(series_provider_id, series_id, name, status):
        """
        Adds a new show with the default settings
        """
        if not find_show(int(series_id), series_provider_id):
            sickrage.app.log.info("Adding show " + str(series_id))
            root_dirs = sickrage.app.config.general.root_dirs.split('|')

            try:
                location = root_dirs[int(root_dirs[0]) + 1]
            except Exception:
                location = None

            if location:
                showPath = os.path.join(location, sanitize_file_name(name))
                dir_exists = make_dir(showPath)

                if not dir_exists:
                    sickrage.app.log.warning("Unable to create the folder %s , can't add the show" % showPath)
                    return
                else:
                    chmod_as_parent(showPath)

                sickrage.app.show_queue.add_show(series_provider_id, int(series_id), showPath,
                                                 default_status=status,
                                                 quality=int(sickrage.app.config.general.quality_default),
                                                 flatten_folders=int(sickrage.app.config.general.flatten_folders_default),
                                                 paused=sickrage.app.config.trakt.start_paused,
                                                 default_status_after=status,
                                                 scene=sickrage.app.config.general.scene_default,
                                                 skip_downloaded=sickrage.app.config.general.skip_downloaded_default)
            else:
                sickrage.app.log.warning(
                    "There was an error creating the show, no root directory setting found")
                return

    def manage_new_show(self, show):
        sickrage.app.log.debug("Checking if trakt watch list wants to search for episodes from new show " + show.name)
        for episode in [i for i in self.todoWanted if i[0] == show.series_id]:
            self.todoWanted.remove(episode)
            set_episode_to_wanted(show, int(episode[1]), int(episode[2]))

    def _check_in_list(self, trakt_id, series_id, season, episode, List=None):
        """
         Check in the Watchlist or CollectionList for Show
         Is the Show, Season and Episode in the trakt_id list (tvdb / tvrage)
        """
        if "Collection" == List:
            try:
                if self.Collectionlist[trakt_id, series_id].seasons[int(season)].episodes[int(episode)]: return True
            except Exception:
                return False
        elif "Show" == List:
            try:
                if self.ShowWatchlist[trakt_id, series_id]: return True
            except Exception:
                return False
        else:
            try:
                if self.EpisodeWatchlist[trakt_id, series_id].seasons[int(season)].episodes[int(episode)]: return True
            except Exception:
                return False

    def _get_show_watchlist(self):
        """
        Get Watchlist and parse once into addressable structure
        """
        try:
            sickrage.app.log.debug("Getting Show Watchlist")
            self.ShowWatchlist = TraktAPI()["sync/watchlist"].shows() or {}
        except Exception as e:
            sickrage.app.log.warning(
                "Could not connect to trakt service, cannot download Show Watchlist: %s" % repr(e))
            return False
        return True

    def _get_episode_watchlist(self):
        """
         Get Watchlist and parse once into addressable structure
        """
        try:
            sickrage.app.log.debug("Getting Episode Watchlist")
            self.EpisodeWatchlist = TraktAPI()["sync/watchlist"].episodes() or {}
        except Exception as e:
            sickrage.app.log.warning(
                "Could not connect to trakt service, cannot download Episode Watchlist: %s" % repr(e))
            return False

        return True

    def _get_show_collection(self):
        """
        Get Collection and parse once into addressable structure
        """
        try:
            sickrage.app.log.debug("Getting Show Collection")
            self.Collectionlist = TraktAPI()["sync/collection"].shows() or {}
        except Exception as e:
            sickrage.app.log.warning(
                "Could not connect to trakt service, cannot download Show Collection: %s" % repr(e))
            return False

        return True

    @staticmethod
    def trakt_bulk_data_generate(data):
        """
        Build the JSON structure to send back to Trakt
        """
        show_list = []

        shows = {}
        seasons = {}

        for series_id, series_provider_id, show_name, startyear, season, episode in data:
            if series_id not in shows:
                shows[series_id] = {'title': show_name,
                                    'year': startyear,
                                    'ids': {sickrage.app.series_providers[series_provider_id].trakt_id: series_id}}

            if series_id not in seasons:
                seasons[series_id] = {}
            if season not in seasons[series_id]:
                seasons[series_id] = {season: []}
            if episode not in seasons[series_id][season]:
                seasons[series_id][season] += [{'number': episode}]

        for series_id, seasonlist in seasons.items():
            if 'seasons' not in shows[series_id]: shows[series_id]['seasons'] = []
            for season, episodelist in seasonlist.items():
                shows[series_id]['seasons'] += [{'number': season, 'episodes': episodelist}]

            show_list.append(shows[series_id])

        return {'shows': show_list}



"""Initial migration

Revision ID: 1
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = '2'
down_revision = '1'


def upgrade():
    with op.batch_alter_table('general') as batch_op:
        batch_op.drop_column('web_host')


def downgrade():
    pass



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################
import os
import re

from tornado.escape import json_decode

import sickrage
from sickrage.core.common import Quality, Qualities, EpisodeStatus
from sickrage.core.databases.main import MainDB
from sickrage.core.databases.main.schemas import IMDbInfoSchema, BlacklistSchema, WhitelistSchema
from sickrage.core.enums import SearchFormat, SeriesProviderID
from sickrage.core.exceptions import CantUpdateShowException, NoNFOException, CantRefreshShowException
from sickrage.core.helpers import checkbox_to_value, sanitize_file_name, make_dir, chmod_as_parent
from sickrage.core.helpers.anidb import short_group_names
from sickrage.core.media.util import series_image, SeriesImageType
from sickrage.core.queues.search import ManualSearchTask
from sickrage.core.tv.episode.helpers import find_episode
from sickrage.core.tv.show.helpers import get_show_list, find_show, find_show_by_slug
from sickrage.core.webserver.handlers.api.v2 import ApiV2BaseHandler
from sickrage.core.websocket import WebSocketMessage
from .schemas import *


class ApiV2SeriesHandler(ApiV2BaseHandler):
    def get(self, series_slug=None):
        """Get list of series or specific series information"
        ---
        tags: [Series]
        summary: Manually search for episodes on search providers
        description: Manually search for episodes on search providers
        parameters:
        - in: path
          schema:
            SeriesSlugPath
        responses:
          200:
            description: Success payload
            content:
              application/json:
                schema:
                  SeriesSuccessSchema
          400:
            description: Bad request; Check `errors` for any validation errors
            content:
              application/json:
                schema:
                  BadRequestSchema
          401:
            description: Returned if your JWT token is missing or expired
            content:
              application/json:
                schema:
                  NotAuthorizedSchema
          404:
            description: Returned if the given series slug does not exist or no series results.
            content:
              application/json:
                schema:
                  NotFoundSchema
        """

        if not series_slug:
            all_series = []

            for show in get_show_list():
                if sickrage.app.show_queue.is_being_removed(show.series_id):
                    continue

                all_series.append(show.to_json(progress=True))

            return self.json_response(all_series)

        series = find_show_by_slug(series_slug)
        if series is None:
            return self._not_found(error=f"Unable to find the specified series using slug: {series_slug}")

        return self.json_response(series.to_json(episodes=True, details=True))

    def post(self):
        data = json_decode(self.request.body)

        is_existing = data.get('isExisting', 'false')

        root_directory = data.get('rootDirectory', None)
        series_id = data.get('seriesId', None)
        series_name = data.get('seriesName', None)
        series_directory = data.get('seriesDirectory', None)
        first_aired = data.get('firstAired', None)
        series_provider_slug = data.get('seriesProviderSlug', None)
        series_provider_language = data.get('seriesProviderLanguage', None)
        default_status = data.get('defaultStatus', None)
        default_status_after = data.get('defaultStatusAfter', None)
        quality_preset = data.get('qualityPreset', None)
        allowed_qualities = data.get('allowedQualities', [])
        preferred_qualities = data.get('preferredQualities', [])
        subtitles = self._parse_boolean(data.get('subtitles', sickrage.app.config.subtitles.default))
        sub_use_sr_metadata = self._parse_boolean(data.get('subUseSrMetadata', 'false'))
        flatten_folders = self._parse_boolean(data.get('flattenFolders', sickrage.app.config.general.flatten_folders_default))
        is_anime = self._parse_boolean(data.get('isAnime', sickrage.app.config.general.anime_default))
        is_scene = self._parse_boolean(data.get('isScene', sickrage.app.config.general.scene_default))
        search_format = data.get('searchFormat', sickrage.app.config.general.search_format_default.name)
        dvd_order = self._parse_boolean(data.get('dvdOrder', 'false'))
        skip_downloaded = self._parse_boolean(data.get('skipDownloaded', sickrage.app.config.general.skip_downloaded_default))
        add_show_year = self._parse_boolean(data.get('addShowYear', 'false'))

        if not series_id:
            return self._bad_request(error=f"Missing seriesId parameter: {series_id}")

        series_provider_id = SeriesProviderID.by_slug(series_provider_slug)
        if not series_provider_id:
            return self._not_found(error="Unable to identify a series provider using provided slug")

        series = find_show(int(series_id), series_provider_id)
        if series:
            return self._bad_request(error=f"Already exists series: {series_id}")

        if is_existing and not series_directory:
            return self._bad_request(error="Missing seriesDirectory parameter")

        if not is_existing:
            series_directory = os.path.join(root_directory, sanitize_file_name(series_name))

            if first_aired:
                series_year = re.search(r'\d{4}', first_aired)
                if add_show_year and not re.match(r'.*\(\d+\)$', series_directory) and series_year:
                    series_directory = f"{series_directory} ({series_year.group()})"

            if os.path.isdir(series_directory):
                sickrage.app.alerts.error(_("Unable to add show"), _("Folder ") + series_directory + _(" exists already"))
                return self._bad_request(error=f"Show directory {series_directory} already exists!")

            if not make_dir(series_directory):
                sickrage.app.log.warning(f"Unable to create the folder {series_directory}, can't add the show")
                sickrage.app.alerts.error(_("Unable to add show"), _(f"Unable to create the folder {series_directory}, can't add the show"))
                return self._bad_request(error=f"Unable to create the show folder {series_directory}, can't add the show")

        chmod_as_parent(series_directory)

        try:
            new_quality = Qualities[quality_preset.upper()]
        except (AttributeError, KeyError):
            new_quality = Quality.combine_qualities([Qualities[x.upper()] for x in allowed_qualities], [Qualities[x.upper()] for x in preferred_qualities])

        sickrage.app.show_queue.add_show(series_provider_id=series_provider_id,
                                         series_id=int(series_id),
                                         showDir=series_directory,
                                         default_status=EpisodeStatus[default_status.upper()],
                                         default_status_after=EpisodeStatus[default_status_after.upper()],
                                         quality=new_quality,
                                         flatten_folders=flatten_folders,
                                         lang=series_provider_language,
                                         subtitles=subtitles,
                                         sub_use_sr_metadata=sub_use_sr_metadata,
                                         anime=is_anime,
                                         dvd_order=dvd_order,
                                         search_format=SearchFormat[search_format.upper()],
                                         paused=False,
                                         # blacklist=blacklist,
                                         # whitelist=whitelist,
                                         scene=is_scene,
                                         skip_downloaded=skip_downloaded)

        sickrage.app.alerts.message(_('Adding Show'), _(f'Adding the specified show into {series_directory}'))

        return self.json_response({'message': True})

    def patch(self, series_slug):
        warnings, errors = [], []

        do_update = False
        do_update_exceptions = False

        data = json_decode(self.request.body)

        series = find_show_by_slug(series_slug)
        if series is None:
            return self._bad_request(error=f"Unable to find the specified series using slug: {series_slug}")

        # if we changed the language then kick off an update
        if data.get('lang') is not None and data['lang'] != series.lang:
            do_update = True

        if data.get('paused') is not None:
            series.paused = checkbox_to_value(data['paused'])

        if data.get('anime') is not None:
            series.anime = checkbox_to_value(data['anime'])

        if data.get('scene') is not None:
            series.scene = checkbox_to_value(data['scene'])

        if data.get('searchFormat') is not None:
            series.search_format = SearchFormat[data['searchFormat']]

        if data.get('subtitles') is not None:
            series.subtitles = checkbox_to_value(data['subtitles'])

        if data.get('subUseSrMetadata') is not None:
            series.sub_use_sr_metadata = checkbox_to_value(data['subUseSrMetadata'])

        if data.get('defaultEpStatus') is not None:
            series.default_ep_status = int(data['defaultEpStatus'])

        if data.get('skipDownloaded') is not None:
            series.skip_downloaded = checkbox_to_value(data['skipDownloaded'])

        if data.get('sceneExceptions') is not None and set(data['sceneExceptions']) != set(series.scene_exceptions):
            do_update_exceptions = True

        if data.get('whitelist') is not None:
            shortwhitelist = short_group_names(data['whitelist'])
            series.release_groups.set_white_keywords(shortwhitelist)

        if data.get('blacklist') is not None:
            shortblacklist = short_group_names(data['blacklist'])
            series.release_groups.set_black_keywords(shortblacklist)

        if data.get('qualityPreset') is not None:
            try:
                new_quality = Qualities[data['qualityPreset']]
            except KeyError:
                new_quality = Quality.combine_qualities([Qualities[x] for x in data['allowedQualities']], [Qualities[x] for x in data['preferredQualities']])

            series.quality = new_quality

        if data.get('flattenFolders') is not None and bool(series.flatten_folders) != bool(data['flattenFolders']):
            series.flatten_folders = data['flattenFolders']
            try:
                sickrage.app.show_queue.refresh_show(series.series_id, series.series_provider_id, True)
            except CantRefreshShowException as e:
                errors.append(_(f"Unable to refresh this show: {e}"))

        if data.get('language') is not None:
            series.lang = data['language']

        if data.get('dvdOrder') is not None:
            series.dvd_order = checkbox_to_value(data['dvdOrder'])

        if data.get('rlsIgnoreWords') is not None:
            series.rls_ignore_words = data['rlsIgnoreWords']

        if data.get('rlsRequireWords') is not None:
            series.rls_require_words = data['rlsRequireWords']

        # series.search_delay = int(data['search_delay'])

        # if we change location clear the db of episodes, change it, write to db, and rescan
        if data.get('location') is not None and os.path.normpath(series.location) != os.path.normpath(data['location']):
            sickrage.app.log.debug(os.path.normpath(series.location) + " != " + os.path.normpath(data['location']))
            if not os.path.isdir(data['location']) and not sickrage.app.config.general.create_missing_show_dirs:
                warnings.append(f"New location {data['location']} does not exist")

            # don't bother if we're going to update anyway
            elif not do_update:
                # change it
                try:
                    series.location = data['location']
                    try:
                        sickrage.app.show_queue.refresh_show(series.series_id, series.series_provider_id, True)
                    except CantRefreshShowException as e:
                        errors.append(_(f"Unable to refresh this show: {e}"))
                        # grab updated info from TVDB
                        # showObj.loadEpisodesFromSeriesProvider()
                        # rescan the episodes in the new folder
                except NoNFOException:
                    warnings.append(_(
                        f"The folder at {data['location']} doesn't contain a tvshow.nfo - copy your files to that folder before you change the directory in SiCKRAGE."))

        # force the update
        if do_update:
            try:
                sickrage.app.show_queue.update_show(series.series_id, series.series_provider_id, force=True)
            except CantUpdateShowException as e:
                errors.append(_(f"Unable to update show: {e}"))

        if do_update_exceptions:
            try:
                series.scene_exceptions = set(data['sceneExceptions'].split(','))
            except CantUpdateShowException:
                warnings.append(_("Unable to force an update on scene exceptions of the show."))

        # if do_update_scene_numbering:
        #     try:
        #         xem_refresh(series.series_id, series.series_provider_id, True)
        #     except CantUpdateShowException:
        #         warnings.append(_("Unable to force an update on scene numbering of the show."))

        # commit changes to database
        series.save()

        return self.json_response(series.to_json(episodes=True, details=True))

    def delete(self, series_slug):
        data = json_decode(self.request.body)

        series = find_show_by_slug(series_slug)
        if series is None:
            return self._not_found(error=f"Unable to find the specified series using slug: {series_slug}")

        sickrage.app.show_queue.remove_show(series.series_id, series.series_provider_id, checkbox_to_value(data.get('delete')))

        return self.json_response({'message': True})


class ApiV2SeriesEpisodesHandler(ApiV2BaseHandler):
    def get(self, series_slug, *args, **kwargs):
        series = find_show_by_slug(series_slug)
        if series is None:
            return self._not_found(error=f"Unable to find the specified series using slug: {series_slug}")

        episodes = []
        for episode in series.episodes:
            episodes.append(episode.to_json())

        return self.json_response(episodes)


class ApiV2SeriesImagesHandler(ApiV2BaseHandler):
    def get(self, series_slug, *args, **kwargs):
        series = find_show_by_slug(series_slug)
        if series is None:
            return self._not_found(error=f"Unable to find the specified series using slug: {series_slug}")

        image = series_image(series.series_id, series.series_provider_id, SeriesImageType.POSTER_THUMB)
        return self.json_response({'poster': image.url})


class ApiV2SeriesImdbInfoHandler(ApiV2BaseHandler):
    def get(self, series_slug, *args, **kwargs):
        series = find_show_by_slug(series_slug)
        if series is None:
            return self._not_found(error=f"Unable to find the specified series using slug: {series_slug}")

        with sickrage.app.main_db.session() as session:
            imdb_info = session.query(MainDB.IMDbInfo).filter_by(imdb_id=series.imdb_id).one_or_none()
            json_data = IMDbInfoSchema().dump(imdb_info)

        return self.json_response(json_data)


class ApiV2SeriesBlacklistHandler(ApiV2BaseHandler):
    def get(self, series_slug, *args, **kwargs):
        series = find_show_by_slug(series_slug)
        if series is None:
            return self._not_found(error=f"Unable to find the specified series using slug: {series_slug}")

        with sickrage.app.main_db.session() as session:
            blacklist = session.query(MainDB.Blacklist).filter_by(series_id=series.series_id, series_provider_id=series.series_provider_id).one_or_none()
            json_data = BlacklistSchema().dump(blacklist)

        return self.json_response(json_data)


class ApiV2SeriesWhitelistHandler(ApiV2BaseHandler):
    def get(self, series_slug, *args, **kwargs):
        series = find_show_by_slug(series_slug)
        if series is None:
            return self._not_found(error=f"Unable to find the specified series using slug: {series_slug}")

        with sickrage.app.main_db.session() as session:
            whitelist = session.query(MainDB.Whitelist).filter_by(series_id=series.series_id, series_provider_id=series.series_provider_id).one_or_none()
            json_data = WhitelistSchema().dump(whitelist)

        return self.json_response(json_data)


class ApiV2SeriesRefreshHandler(ApiV2BaseHandler):
    def get(self, series_slug):
        force = self.get_argument('force', None)

        series = find_show_by_slug(series_slug)
        if series is None:
            return self._not_found(error=f"Unable to find the specified series using slug: {series_slug}")

        try:
            sickrage.app.show_queue.refresh_show(series.series_id, series.series_provider_id, force=bool(force))
        except CantUpdateShowException as e:
            return self._bad_request(error=_(f"Unable to refresh this show, error: {e}"))


class ApiV2SeriesUpdateHandler(ApiV2BaseHandler):
    def get(self, series_slug):
        force = self.get_argument('force', None)

        series = find_show_by_slug(series_slug)
        if series is None:
            return self._not_found(error=f"Unable to find the specified series using slug: {series_slug}")

        try:
            sickrage.app.show_queue.update_show(series.series_id, series.series_provider_id, force=bool(force))
        except CantUpdateShowException as e:
            return self._bad_request(error=_(f"Unable to update this show, error: {e}"))


class ApiV2SeriesEpisodesRenameHandler(ApiV2BaseHandler):
    def get(self, series_slug):
        """Get list of episodes to rename"
        ---
        tags: [Series]
        summary: Get list of episodes to rename
        description: Get list of episodes to rename
        parameters:
        - in: path
          schema:
            SeriesSlugPath
        responses:
          200:
            description: Success payload
            content:
              application/json:
                schema:
                  SeriesEpisodesRenameSuccessSchema
          400:
            description: Bad request; Check `errors` for any validation errors
            content:
              application/json:
                schema:
                  BadRequestSchema
          401:
            description: Returned if your JWT token is missing or expired
            content:
              application/json:
                schema:
                  NotAuthorizedSchema
        """
        if not series_slug:
            return self._bad_request(error="Missing series slug")

        rename_data = []

        series = find_show_by_slug(series_slug)
        if series is None:
            return self._not_found(error=f"Unable to find the specified series using slug: {series_slug}")

        if not os.path.isdir(series.location):
            return self._bad_request(error="Can't rename episodes when the show location does not exist")

        for episode in series.episodes:
            if not episode.location:
                continue

            current_location = episode.location[len(episode.show.location) + 1:]
            new_location = "{}.{}".format(episode.proper_path(), current_location.split('.')[-1])

            if current_location != new_location:
                rename_data.append({
                    'episodeId': episode.episode_id,
                    'season': episode.season,
                    'episode': episode.episode,
                    'currentLocation': current_location,
                    'newLocation': new_location,
                })

        return self.json_response(rename_data)

    def post(self, series_slug):
        """Rename list of episodes"
        ---
        tags: [Series]
        summary: Rename list of episodes
        description: Rename list of episodes
        parameters:
        - in: path
          schema:
            SeriesSlugPath
        responses:
          200:
            description: Success payload
            content:
              application/json:
                schema:
                  EpisodesRenameSuccessSchema
          400:
            description: Bad request; Check `errors` for any validation errors
            content:
              application/json:
                schema:
                  BadRequestSchema
          401:
            description: Returned if your JWT token is missing or expired
            content:
              application/json:
                schema:
                  NotAuthorizedSchema
        """
        data = json_decode(self.request.body)

        renamed_episodes = []

        series = find_show_by_slug(series_slug)
        if series is None:
            return self._not_found(error=f"Unable to find the specified series using slug: {series_slug}")

        if not os.path.isdir(series.location):
            return self._bad_request(error="Can't rename episodes when the show location does not exist")

        for episode_id in data.get('episodeIdList', []):
            episode = find_episode(episode_id, series.series_provider_id)
            if episode:
                episode.rename()
                renamed_episodes.append(episode.episode_id)

        if len(renamed_episodes) > 0:
            WebSocketMessage('SHOW_RENAMED', {'seriesSlug': series.slug}).push()

        return self.json_response(renamed_episodes)


class ApiV2SeriesEpisodesManualSearchHandler(ApiV2BaseHandler):
    def get(self, series_slug, episode_slug):
        """Episode Manual Search"
        ---
        tags: [Series]
        summary: Manually search for episode on search providers
        description: Manually search for episode on search providers
        parameters:
        - in: path
          schema:
            SeriesSlugPath
        - in: path
          schema:
            EpisodeSlugPath
        responses:
          200:
            description: Success payload
            content:
              application/json:
                schema:
                  EpisodesManualSearchSuccessSchema
          400:
            description: Bad request; Check `errors` for any validation errors
            content:
              application/json:
                schema:
                  BadRequestSchema
          401:
            description: Returned if your JWT token is missing or expired
            content:
              application/json:
                schema:
                  NotAuthorizedSchema
          404:
            description: Returned if the given episode slug does not exist or the search returns no results.
            content:
              application/json:
                schema:
                  NotFoundSchema
        """
        use_existing_quality = self.get_argument('useExistingQuality', None) or False

        # validation_errors = self._validate_schema(SeriesEpisodesManualSearchPath, self.request.path)
        # if validation_errors:
        #     return self._bad_request(error=validation_errors)
        #
        # validation_errors = self._validate_schema(SeriesEpisodesManualSearchSchema, self.request.arguments)
        # if validation_errors:
        #     return self._bad_request(error=validation_errors)
        #

        series = find_show_by_slug(series_slug)
        if series is None:
            return self._not_found(error=f"Unable to find the specified series using slug: {series_slug}")

        match = re.match(r'^s(?P<season>\d+)e(?P<episode>\d+)$', episode_slug)
        season_num = match.group('season')
        episode_num = match.group('episode')

        episode = series.get_episode(int(season_num), int(episode_num), no_create=True)
        if episode is None:
            return self._bad_request(error=f"Unable to find the specified episode using slug: {episode_slug}")

        # make a queue item for it and put it on the queue
        ep_queue_item = ManualSearchTask(int(episode.show.series_id),
                                         episode.show.series_provider_id,
                                         int(episode.season),
                                         int(episode.episode),
                                         bool(use_existing_quality))

        sickrage.app.search_queue.put(ep_queue_item)
        if not all([ep_queue_item.started, ep_queue_item.success]):
            return self.json_response({'success': True})

        return self._not_found(error=_(f"Unable to find season {season_num} episode {episode_num} for show {series.name} on search providers"))



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################

import sickrage
from sickrage.core.exceptions import FailedPostProcessingFailedException, EpisodeNotFoundException
from sickrage.core.helpers import show_names
from sickrage.core.nameparser import InvalidNameException, InvalidShowException, NameParser
from sickrage.core.common import Quality, EpisodeStatus
from sickrage.core.queues.search import FailedSearchTask
from sickrage.core.tv.show.helpers import find_show


class FailedProcessor(object):
    """Take appropriate action when a download fails to complete"""

    def __init__(self, dirName, nzbName):
        """
        :param dirName: Full path to the folder of the failed download
        :param nzbName: Full name of the nzb file that failed
        """
        self.dir_name = dirName
        self.nzb_name = nzbName

        self.log = ""

    def process(self):
        """
        Do the actual work

        :return: True
        """
        self._log("Failed download detected: (" + str(self.nzb_name) + ", " + str(self.dir_name) + ")")

        release_name = show_names.determine_release_name(self.dir_name, self.nzb_name)
        if release_name is None:
            self._log("Warning: unable to find a valid release name.", sickrage.app.log.WARNING)
            raise FailedPostProcessingFailedException()

        try:
            parsed = NameParser(False).parse(release_name)
        except InvalidNameException:
            self._log("Error: release name is invalid: " + release_name, sickrage.app.log.DEBUG)
            raise FailedPostProcessingFailedException()
        except InvalidShowException:
            self._log("Error: unable to parse release name " + release_name + " into a valid show",
                      sickrage.app.log.DEBUG)
            raise FailedPostProcessingFailedException()

        series = find_show(parsed.series_id, parsed.series_provider_id)
        if not series:
            raise FailedPostProcessingFailedException()

        if series.paused:
            self._log("Warning: skipping failed processing for {} because the show is paused".format(release_name), sickrage.app.log.DEBUG)
            raise FailedPostProcessingFailedException()

        sickrage.app.log.debug("name_parser info: ")
        sickrage.app.log.debug(" - " + str(parsed.series_name))
        sickrage.app.log.debug(" - " + str(parsed.season_number))
        sickrage.app.log.debug(" - " + str(parsed.episode_numbers))
        sickrage.app.log.debug(" - " + str(parsed.extra_info))
        sickrage.app.log.debug(" - " + str(parsed.release_group))
        sickrage.app.log.debug(" - " + str(parsed.air_date))

        for episode in parsed.episode_numbers:
            try:
                episode_obj = series.get_episode(parsed.season_number, episode)
            except EpisodeNotFoundException as e:
                continue

            cur_status, cur_quality = Quality.split_composite_status(episode_obj.status)
            if cur_status not in (EpisodeStatus.SNATCHED, EpisodeStatus.SNATCHED_BEST, EpisodeStatus.SNATCHED_PROPER):
                continue

            sickrage.app.search_queue.put(FailedSearchTask(parsed.series_id, parsed.series_provider_id, episode_obj.season, episode_obj.episode))

        return True

    def _log(self, message, level=None):
        """Log to regular logfile and save for return for PP script log"""
        sickrage.app.log.log(level or sickrage.app.log.INFO, message)
        self.log += message + "\n"
